#!/usr/bin/env python3
"""Functionality for data/records.bib and git repository."""
from __future__ import annotations

import os
import tempfile
import time
import typing
from pathlib import Path
from random import randint

import git
from git import GitCommandError
from git import InvalidGitRepositoryError

import colrev.exceptions as colrev_exceptions
import colrev.loader.bib
import colrev.loader.load_utils
import colrev.ops.check
import colrev.process.operation
import colrev.record.record
import colrev.record.record_id_setter
import colrev.record.record_prep
from colrev.constants import ExitCodes
from colrev.constants import Fields
from colrev.constants import FileSets
from colrev.constants import RecordState
from colrev.writer.write_utils import to_string

# pylint: disable=too-many-public-methods


class Dataset:
    """The CoLRev dataset (records and their history in git)"""

    _git_repo: git.Repo

    def __init__(self, *, review_manager: colrev.review_manager.ReviewManager) -> None:
        self.review_manager = review_manager

        try:
            # In most cases, the repo should exist
            # due to review_manager._get_project_home_dir()
            self._git_repo = git.Repo(self.review_manager.path)
        except InvalidGitRepositoryError as exc:
            msg = "Not a CoLRev/git repository. Run\n    colrev init"
            raise colrev_exceptions.RepoSetupError(msg) from exc

        self.update_gitignore(
            add=FileSets.DEFAULT_GIT_IGNORE_ITEMS,
            remove=FileSets.DEPRECATED_GIT_IGNORE_ITEMS,
        )

    def update_gitignore(
        self, *, add: typing.Optional[list] = None, remove: typing.Optional[list] = None
    ) -> None:
        """Update the gitignore file by adding or removing particular paths"""
        # The following may print warnings...

        git_ignore_file = self.review_manager.paths.git_ignore
        if git_ignore_file.is_file():
            gitignore_content = git_ignore_file.read_text(encoding="utf-8")
        else:
            gitignore_content = ""
        ignored_items = gitignore_content.splitlines()
        if remove:
            ignored_items = [x for x in ignored_items if x not in remove]
        if add:
            ignored_items = ignored_items + [
                str(a) for a in add if str(a) not in ignored_items
            ]

        with git_ignore_file.open("w", encoding="utf-8") as file:
            file.write("\n".join(ignored_items) + "\n")
        self.add_changes(git_ignore_file)

    def get_origin_state_dict(self, records_string: str = "") -> dict:
        """Get the origin_state_dict (to determine state transitions efficiently)

        {'30_example_records.bib/Staehr2010': <RecordState.pdf_not_available: 10>,}
        """

        current_origin_states_dict = {}
        if records_string != "":
            with tempfile.NamedTemporaryFile(
                mode="wb", delete=False, suffix=".bib"
            ) as temp_file:
                temp_file.write(records_string.encode("utf-8"))
                temp_file_path = Path(temp_file.name)
            bib_loader = colrev.loader.bib.BIBLoader(
                filename=temp_file_path,
                logger=self.review_manager.logger,
                unique_id_field="ID",
            )
        else:
            bib_loader = colrev.loader.bib.BIBLoader(
                filename=self.review_manager.paths.records,
                logger=self.review_manager.logger,
                unique_id_field="ID",
            )
        for record_header_item in bib_loader.get_record_header_items().values():
            for origin in record_header_item[Fields.ORIGIN]:
                current_origin_states_dict[origin] = record_header_item[Fields.STATUS]
        return current_origin_states_dict

    def get_committed_origin_state_dict(self) -> dict:
        """Get the committed origin_state_dict"""
        revlist = (
            (
                commit.hexsha,
                (
                    commit.tree / self.review_manager.paths.RECORDS_FILE_GIT
                ).data_stream.read(),
            )
            for commit in self._git_repo.iter_commits(
                paths=self.review_manager.paths.RECORDS_FILE_GIT
            )
        )
        filecontents = list(revlist)[0][1]

        committed_origin_state_dict = self.get_origin_state_dict(
            filecontents.decode("utf-8")
        )
        return committed_origin_state_dict

    def load_records_from_history(self, commit_sha: str = "") -> typing.Iterator[dict]:
        """
        Iterates through Git history, yielding records file contents as dictionaries.

        Starts iteration from a provided commit SHA.
        Skips commits where the records file is unchanged.
        Useful for tracking dataset changes over time.

        Parameters:
            commit_sha (str, optional): Start iteration from this commit SHA.
            Defaults to beginning of Git history if not provided.

        Yields:
            dict: Records file contents at a specific Git history point, as a dictionary.
        """

        reached_target_commit = False  # if no commit_sha provided
        for current_commit in self._git_repo.iter_commits(
            paths=self.review_manager.paths.RECORDS_FILE_GIT
        ):

            # Skip all commits before the specified commit_sha, if provided
            if commit_sha and not reached_target_commit:
                if commit_sha != current_commit.hexsha:
                    # Move to the next commit
                    continue
                reached_target_commit = True

            # Read and parse the records file from the current commit
            filecontents = (
                current_commit.tree / self.review_manager.paths.RECORDS_FILE_GIT
            ).data_stream.read()

            records_dict = colrev.loader.load_utils.loads(
                load_string=filecontents.decode("utf-8", "replace"),
                implementation="bib",
                logger=self.review_manager.logger,
            )
            if records_dict:
                yield records_dict

    def load_records_dict(
        self,
        *,
        header_only: bool = False,
    ) -> dict:
        """Load the records

        header_only:

        {"Staehr2010": {'ID': 'Staehr2010',
        'colrev_origin': ['30_example_records.bib/Staehr2010'],
        'colrev_status': <RecordState.md_imported: 2>,
        'screening_criteria': 'criterion1=in;criterion2=out',
        'file': PosixPath('data/pdfs/Smith2000.pdf'),
        'colrev_data_provenance': {Fields.AUTHOR:{"source":"...", "note":"..."}}},
        }
        """

        if self.review_manager.notified_next_operation is None:
            raise colrev_exceptions.ReviewManagerNotNotifiedError()

        if header_only:
            # Note : currently not parsing screening_criteria to settings.ScreeningCriterion
            # to optimize performance
            bib_loader = colrev.loader.bib.BIBLoader(
                filename=self.review_manager.paths.records,
                logger=self.review_manager.logger,
                unique_id_field="ID",
            )
            return bib_loader.get_record_header_items()

        if self.review_manager.paths.records.is_file():

            records_dict = colrev.loader.load_utils.load(
                filename=self.review_manager.paths.records,
                logger=self.review_manager.logger,
                unique_id_field="ID",
            )

        else:
            records_dict = {}

        return records_dict

    def save_records_dict_to_file(self, records: dict) -> None:
        """Save the records dict"""
        # Note : this classmethod function can be called by CoLRev scripts
        # operating outside a CoLRev repo (e.g., sync)

        bibtex_str = to_string(records_dict=records, implementation="bib")

        with open(self.review_manager.paths.records, "w", encoding="utf-8") as out:
            out.write(bibtex_str + "\n")

        self._add_record_changes()

    def _save_record_list_by_id(self, records: dict) -> None:

        parsed = to_string(records_dict=records, implementation="bib")
        record_list = [
            {
                Fields.ID: item[item.find("{") + 1 : item.find(",")],
                "record": "@" + item + "\n",
            }
            for item in parsed.split("\n@")
        ]
        # Correct the first item
        record_list[0]["record"] = "@" + record_list[0]["record"][2:]

        current_id_str = "NOTSET"
        if self.review_manager.paths.records.is_file():
            with open(self.review_manager.paths.records, "r+b") as file:
                seekpos = file.tell()
                line = file.readline()
                while line:
                    if b"@" in line[:3]:
                        current_id = line[line.find(b"{") + 1 : line.rfind(b",")]
                        current_id_str = current_id.decode("utf-8")
                    if current_id_str in [x[Fields.ID] for x in record_list]:
                        replacement = [
                            x["record"]
                            for x in record_list
                            if x[Fields.ID] == current_id_str
                        ][0]
                        record_list = [
                            x for x in record_list if x[Fields.ID] != current_id_str
                        ]
                        line = file.readline()
                        while (
                            b"@" not in line[:3] and line
                        ):  # replace: drop the current record
                            line = file.readline()
                        remaining = line + file.read()
                        file.seek(seekpos)
                        file.write(replacement.encode("utf-8"))
                        seekpos = file.tell()
                        file.flush()
                        os.fsync(file)
                        file.write(remaining)
                        file.truncate()  # if the replacement is shorter...
                        file.seek(seekpos)

                    seekpos = file.tell()
                    line = file.readline()

        if len(record_list) > 0:
            with open(
                self.review_manager.paths.records, "a", encoding="utf8"
            ) as m_refs:
                for item in record_list:
                    m_refs.write(item["record"])

        self._add_record_changes()

    def save_records_dict(self, records: dict, *, partial: bool = False) -> None:
        """Save the records dict in RECORDS_FILE"""

        if partial:
            self._save_record_list_by_id(records)
            return
        self.save_records_dict_to_file(records)

    def read_next_record(self, *, conditions: list) -> typing.Iterator[dict]:
        """Read records (Iterator) based on condition"""

        # Note : matches conditions connected with 'OR'
        records = self.load_records_dict()

        records_list = []
        for _, record in records.items():
            for condition in conditions:
                for key, value in condition.items():
                    if str(value) == str(record[key]):
                        records_list.append(record)
        yield from records_list

    def format_records_file(self) -> dict:
        """Format the records file (Entrypoint for pre-commit hooks)"""

        if (
            not self.review_manager.paths.records.is_file()
            or not self.records_changed()
        ):
            return {"status": ExitCodes.SUCCESS, "msg": "Everything ok."}

        colrev.ops.check.CheckOperation(self.review_manager)  # to notify
        quality_model = self.review_manager.get_qm()
        records = self.load_records_dict()
        for record_dict in records.values():
            if Fields.STATUS not in record_dict:
                return {
                    "status": ExitCodes.FAIL,
                    "msg": f" no status field in record ({record_dict[Fields.ID]})",
                }

            record = colrev.record.record_prep.PrepRecord(record_dict)
            if record_dict[Fields.STATUS] in [
                RecordState.md_needs_manual_preparation,
            ]:
                record.run_quality_model(quality_model, set_prepared=True)

            if record_dict[Fields.STATUS] == RecordState.pdf_prepared:
                record.reset_pdf_provenance_notes()

        self.save_records_dict(records)
        changed = self.review_manager.paths.RECORDS_FILE in [
            r.a_path for r in self._git_repo.index.diff(None)
        ]
        self.review_manager.update_status_yaml()
        self.review_manager.load_settings()
        self.review_manager.save_settings()

        if changed:  # pragma: no cover
            return {"status": ExitCodes.FAIL, "msg": "Records formatted"}

        return {"status": ExitCodes.SUCCESS, "msg": "Everything ok."}

    # ID creation, update and lookup ---------------------------------------

    def propagated_id(self, *, record_id: str) -> bool:
        """Check whether an ID is propagated (i.e., its record's status is beyond md_processed)"""

        for record in self.load_records_dict(header_only=True).values():
            if record[Fields.ID] == record_id:
                if record[Fields.STATUS] in RecordState.get_post_x_states(
                    state=RecordState.md_processed
                ):
                    return True

        return False

    def set_ids(self, selected_ids: typing.Optional[list] = None) -> dict:
        """Set the IDs of records according to predefined formats or
        according to the LocalIndex"""
        id_setter = colrev.record.record_id_setter.IDSetter(
            id_pattern=self.review_manager.settings.project.id_pattern,
            skip_local_index=self.review_manager.settings.is_curated_masterdata_repo(),
        )
        records = self.load_records_dict()
        updated_records = id_setter.set_ids(
            records=records,
            selected_ids=selected_ids,
        )
        self.save_records_dict(records)
        self.add_changes(self.review_manager.paths.RECORDS_FILE)
        return updated_records

    # GIT operations -----------------------------------------------

    def get_repo(self) -> git.Repo:
        """Get the git repository object"""

        if self.review_manager.notified_next_operation is None:
            raise colrev_exceptions.ReviewManagerNotNotifiedError()
        return self._git_repo

    def repo_initialized(self) -> bool:
        """Check whether the repository is initialized"""
        try:
            self._git_repo.head.commit
        except ValueError:
            return False
        return True

    def has_record_changes(self, *, change_type: str = "all") -> bool:
        """Check whether the records have changes"""
        return self.has_changes(
            Path(self.review_manager.paths.RECORDS_FILE_GIT), change_type=change_type
        )

    def has_changes(self, relative_path: Path, *, change_type: str = "all") -> bool:
        """Check whether the relative path (or the git repository) has changes"""

        assert change_type in [
            "all",
            "staged",
            "unstaged",
        ], "Invalid change_type specified"

        # Check if the repository has at least one commit
        try:
            bool(self._git_repo.head.commit)
        except ValueError:
            return True  # Repository has no commit

        diff_index = [item.a_path for item in self._git_repo.index.diff(None)]
        diff_head = [item.a_path for item in self._git_repo.head.commit.diff()]
        unstaged_changes = diff_index + self._git_repo.untracked_files

        # Ensure the path uses forward slashes, which is compatible with Git's path handling
        path_str = str(relative_path).replace("\\", "/")

        if change_type == "all":
            path_changed = path_str in diff_index + diff_head
        elif change_type == "staged":
            path_changed = path_str in diff_head
        elif change_type == "unstaged":
            path_changed = path_str in unstaged_changes
        else:
            return False
        return path_changed

    def _sleep_util_git_unlocked(self) -> None:
        i = 0
        while (
            self.review_manager.path / Path(".git/index.lock")
        ).is_file():  # pragma: no cover
            i += 1
            time.sleep(randint(1, 50) * 0.1)  # nosec
            if i > 5:
                print("Waiting for previous git operation to complete")
            elif i > 30:
                raise colrev_exceptions.GitNotAvailableError()

    def add_changes(
        self, path: Path, *, remove: bool = False, ignore_missing: bool = False
    ) -> None:
        """Add changed file to git"""

        if path.is_absolute():
            path = path.relative_to(self.review_manager.path)
        path_str = str(path).replace("\\", "/")

        self._sleep_util_git_unlocked()
        try:
            if remove:
                self._git_repo.index.remove([path_str])
            else:
                self._git_repo.index.add([path_str])
        except FileNotFoundError as exc:
            if not ignore_missing:
                raise exc

    def get_untracked_files(self) -> list:
        """Get the files that are untracked by git"""

        return [Path(x) for x in self._git_repo.untracked_files]

    def records_changed(self) -> bool:
        """Check whether the records were changed"""
        main_recs_changed = self.review_manager.paths.RECORDS_FILE_GIT in [
            item.a_path for item in self._git_repo.index.diff(None)
        ] + [x.a_path for x in self._git_repo.head.commit.diff()]
        return main_recs_changed

    # pylint: disable=too-many-arguments
    def create_commit(
        self,
        *,
        msg: str,
        manual_author: bool = False,
        script_call: str = "",
        saved_args: typing.Optional[dict] = None,
        skip_status_yaml: bool = False,
        skip_hooks: bool = True,
    ) -> bool:
        """Create a commit (including a commit report)"""
        # pylint: disable=import-outside-toplevel
        # pylint: disable=redefined-outer-name
        import colrev.ops.commit

        if self.review_manager.exact_call and script_call == "":
            script_call = self.review_manager.exact_call

        commit = colrev.ops.commit.Commit(
            review_manager=self.review_manager,
            msg=msg,
            manual_author=manual_author,
            script_name=script_call,
            saved_args=saved_args,
            skip_hooks=skip_hooks,
        )
        ret = commit.create(skip_status_yaml=skip_status_yaml)
        return ret

    def file_in_history(self, filepath: Path) -> bool:
        """Check whether a file is in the git history"""
        return str(filepath) in [
            o.path for o in self._git_repo.head.commit.tree.traverse()
        ]

    def get_commit_message(self, *, commit_nr: int) -> str:
        """Get the commit message for commit #"""
        master = self._git_repo.head.reference
        assert commit_nr == 0  # extension : implement other cases
        if commit_nr == 0:
            cmsg = master.commit.message
            return cmsg
        return ""

    def _add_record_changes(self) -> None:
        """Add changes in records to git"""
        self._sleep_util_git_unlocked()
        self._git_repo.index.add([str(self.review_manager.paths.RECORDS_FILE)])

    def add_setting_changes(self) -> None:
        """Add changes in settings to git"""
        self._sleep_util_git_unlocked()

        self._git_repo.index.add([str(self.review_manager.paths.SETTINGS_FILE)])

    def has_untracked_search_records(self) -> bool:
        """Check whether there are untracked search records"""
        return any(
            str(self.review_manager.paths.SEARCH_DIR) in str(untracked_file)
            for untracked_file in self.get_untracked_files()
        )

    def stash_unstaged_changes(self) -> bool:
        """Stash unstaged changes"""
        ret = self._git_repo.git.stash("push", "--keep-index")
        return "No local changes to save" != ret

    def reset_log_if_no_changes(self) -> None:
        """Reset the report log file if there are not changes"""
        if not self._git_repo.is_dirty():
            self.review_manager.reset_report_logger()

    def get_last_commit_sha(self) -> str:  # pragma: no cover
        """Get the last commit sha"""
        return str(self._git_repo.head.commit.hexsha)

    def get_tree_hash(self) -> str:  # pragma: no cover
        """Get the current tree hash"""
        tree_hash = self._git_repo.git.execute(["git", "write-tree"])
        return str(tree_hash)

    def _get_remote_commit_differences(self) -> list:  # pragma: no cover
        origin = self._git_repo.remotes.origin
        if origin.exists():
            try:
                origin.fetch()
            except GitCommandError:
                return [-1, -1]

        nr_commits_behind, nr_commits_ahead = -1, -1
        if self._git_repo.active_branch.tracking_branch() is not None:
            branch_name = str(self._git_repo.active_branch)
            tracking_branch_name = str(self._git_repo.active_branch.tracking_branch())
            # self.review_manager.logger.debug(f"{branch_name} - {tracking_branch_name}")

            behind_operation = branch_name + ".." + tracking_branch_name
            commits_behind = self._git_repo.iter_commits(behind_operation)
            nr_commits_behind = sum(1 for c in commits_behind)

            ahead_operation = tracking_branch_name + ".." + branch_name
            commits_ahead = self._git_repo.iter_commits(ahead_operation)
            nr_commits_ahead = sum(1 for c in commits_ahead)

        return [nr_commits_behind, nr_commits_ahead]

    def behind_remote(self) -> bool:  # pragma: no cover
        """Check whether the repository is behind the remote"""
        nr_commits_behind = 0
        connected_remote = 0 != len(self._git_repo.remotes)
        if connected_remote:
            origin = self._git_repo.remotes.origin
            if origin.exists():
                (
                    nr_commits_behind,
                    _,
                ) = self._get_remote_commit_differences()
        if nr_commits_behind > 0:
            return True
        return False

    def remote_ahead(self) -> bool:  # pragma: no cover
        """Check whether the remote is ahead"""
        connected_remote = 0 != len(self._git_repo.remotes)
        if connected_remote:
            origin = self._git_repo.remotes.origin
            if origin.exists():
                (
                    _,
                    nr_commits_ahead,
                ) = self._get_remote_commit_differences()
                if nr_commits_ahead > 0:
                    return True
        return False

    def pull_if_repo_clean(self) -> None:  # pragma: no cover
        """Pull project if repository is clean"""
        if not self._git_repo.is_dirty():
            origin = self._git_repo.remotes.origin
            origin.pull()

    def get_remote_url(self) -> str:  # pragma: no cover
        """Get the remote url"""
        remote_url = "NA"
        for remote in self._git_repo.remotes:
            if remote.name == "origin":
                remote_url = remote.url
        return remote_url

    def get_last_commit_date(self, filename: Path) -> str:
        """Get the last commit date for a file"""
        last_commit = next(self._git_repo.iter_commits(paths=filename))
        return last_commit.committed_datetime.isoformat()



#! /usr/bin/env python
"""SearchSource: Scopus"""
from __future__ import annotations

from pathlib import Path

import zope.interface
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class ScopusSearchSource:
    """Scopus"""

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings
    endpoint = "colrev.scopus"
    # pylint: disable=colrev-missed-constant-usage
    source_identifier = "url"
    search_types = [SearchType.DB]

    ci_supported: bool = Field(default=False)
    heuristic_status = SearchSourceHeuristicStatus.supported

    db_url = "https://www.scopus.com/search/form.uri?display=advanced"

    def __init__(
        self, *, source_operation: colrev.process.operation.Operation, settings: dict
    ) -> None:
        self.review_manager = source_operation.review_manager
        self.search_source = self.settings_class(**settings)
        self.quality_model = self.review_manager.get_qm()
        self.operation = source_operation

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for Scopus"""

        result = {"confidence": 0.0}
        if "source={Scopus}," in data:
            result["confidence"] = 1.0
            return result

        if "www.scopus.com" in data:
            if data.count("www.scopus.com") >= data.count("\n@"):
                result["confidence"] = 1.0

        return result

    @classmethod
    def add_endpoint(
        cls,
        operation: colrev.ops.search.Search,
        params: str,
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint (based on query provided to colrev search --add )"""

        search_source = operation.create_db_source(
            search_source_cls=cls,
            params={},
        )
        operation.add_source_and_search(search_source)
        return search_source

    def search(self, rerun: bool) -> None:
        """Run a search of Scopus"""

        if self.search_source.search_type == SearchType.DB:
            self.operation.run_db_search(  # type: ignore
                search_source_cls=self.__class__,
                source=self.search_source,
            )
            return

        raise NotImplementedError

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.record.Record:
        """Not implemented"""
        return record

    def _load_bib(self) -> dict:

        def entrytype_setter(record_dict: dict) -> None:
            if "document_type" in record_dict:
                if record_dict["document_type"] == "Conference Paper":
                    record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.INPROCEEDINGS

                elif record_dict["document_type"] == "Conference Review":
                    record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.PROCEEDINGS

                elif record_dict["document_type"] == "Article":
                    record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.ARTICLE

        def field_mapper(record_dict: dict) -> None:

            if record_dict[Fields.ENTRYTYPE] in [
                ENTRYTYPES.INPROCEEDINGS,
                ENTRYTYPES.PROCEEDINGS,
            ]:
                record_dict[Fields.BOOKTITLE] = record_dict.pop(Fields.JOURNAL, None)

            if record_dict[Fields.ENTRYTYPE] == ENTRYTYPES.BOOK:
                if (
                    Fields.JOURNAL in record_dict
                    and Fields.BOOKTITLE not in record_dict
                ):
                    record_dict[Fields.BOOKTITLE] = record_dict.pop(Fields.TITLE, None)
                    record_dict[Fields.TITLE] = record_dict.pop(Fields.JOURNAL, None)

            if "art_number" in record_dict:
                record_dict[f"{self.endpoint}.art_number"] = record_dict.pop(
                    "art_number"
                )
            if "note" in record_dict:
                record_dict[f"{self.endpoint}.note"] = record_dict.pop("note")
            if "document_type" in record_dict:
                record_dict[f"{self.endpoint}.document_type"] = record_dict.pop(
                    "document_type"
                )
            if "source" in record_dict:
                record_dict[f"{self.endpoint}.source"] = record_dict.pop("source")

            if "Start_Page" in record_dict and "End_Page" in record_dict:
                if (
                    record_dict["Start_Page"] != "nan"
                    and record_dict["End_Page"] != "nan"
                ):
                    record_dict[Fields.PAGES] = (
                        record_dict["Start_Page"] + "--" + record_dict["End_Page"]
                    )
                    record_dict[Fields.PAGES] = record_dict[Fields.PAGES].replace(
                        ".0", ""
                    )
                    del record_dict["Start_Page"]
                    del record_dict["End_Page"]

        records = colrev.loader.load_utils.load(
            filename=self.search_source.filename,
            unique_id_field="ID",
            entrytype_setter=entrytype_setter,
            field_mapper=field_mapper,
            logger=self.review_manager.logger,
        )

        return records

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":
            return self._load_bib()

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.record.Record:
        """Source-specific preparation for Scopus"""

        return record



#! /usr/bin/env python
"""Functionality to identify records."""
from __future__ import annotations

import logging
import os
import re
import tempfile
import typing
from pathlib import Path

import imagehash
import pymupdf
from nameparser import HumanName
from PIL import Image

import colrev.env.utils
import colrev.exceptions as colrev_exceptions
from colrev.constants import Colors
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields
from colrev.constants import FieldSet
from colrev.constants import FieldValues
from colrev.constants import RecordState

if typing.TYPE_CHECKING:  # pragma: no cover
    import colrev.record.record


def _format_author_field_for_cid(input_string: str) -> str:
    input_string = input_string.replace("\n", " ").replace("'", "")
    names = input_string.replace("; ", " and ").split(" and ")
    author_list = []
    for name in names:
        if name.rstrip()[-1:] == ",":
            # if last-names only (eg, "Webster, and Watson, ")
            if len(name[:-2]) > 1:
                author_list.append(str(name.rstrip()[:-1]))
        else:
            parsed_name = HumanName(name)
            # Note: do not set parsed_name.string_format as a global constant
            # to preserve consistent creation of identifiers
            parsed_name.string_format = "{last} "
            if len(parsed_name.middle) > 0:
                parsed_name.middle = parsed_name.middle[:1]
            if len(parsed_name.first) > 0:
                parsed_name.first = parsed_name.first[:1]
            if len(parsed_name.nickname) > 0:
                parsed_name.nickname = ""

            if len(str(parsed_name)) > 1:
                author_list.append(str(parsed_name))

    return " ".join(author_list)


def _get_container_title(record: colrev.record.record.Record) -> str:
    # Note: custom __get_container_title for the colrev_id

    # school as the container title for theses
    if record.data[Fields.ENTRYTYPE] in ["phdthesis", "masterthesis"]:
        container_title = record.data[Fields.SCHOOL]
    # for technical reports
    elif record.data[Fields.ENTRYTYPE] == ENTRYTYPES.TECHREPORT:
        container_title = record.data["institution"]
    elif record.data[Fields.ENTRYTYPE] == ENTRYTYPES.INPROCEEDINGS:
        container_title = record.data[Fields.BOOKTITLE]
    elif record.data[Fields.ENTRYTYPE] == ENTRYTYPES.ARTICLE:
        container_title = record.data[Fields.JOURNAL]
    elif Fields.SERIES in record.data:
        container_title = record.data[Fields.SERIES]
    elif Fields.URL in record.data:
        container_title = record.data[Fields.URL]
    else:  # pragma: no cover
        raise KeyError

    return container_title


def _robust_append(input_string: str, *, to_append: str) -> str:
    input_string = str(input_string)
    to_append = str(to_append).replace("\n", " ").replace("/", " ")
    to_append = to_append.rstrip().lstrip().replace("–", " ")
    to_append = to_append.replace("emph{", "")
    to_append = to_append.replace("&amp;", "and")
    to_append = to_append.replace(" & ", " and ")
    to_append = colrev.env.utils.remove_accents(to_append)
    to_append = re.sub("[^0-9a-zA-Z -]+", "", to_append)
    to_append = re.sub(r"\s+", "-", to_append)
    to_append = re.sub(r"-+", "-", to_append)
    to_append = to_append.lower()
    if len(to_append) > 1:
        to_append = to_append.rstrip("-")
    input_string = input_string + "|" + to_append
    return input_string


def _check_colrev_id_preconditions(
    record: colrev.record.record.Record,
    *,
    assume_complete: bool,
) -> None:
    if assume_complete:
        return
    if record.data.get(Fields.STATUS, "NA") in [
        RecordState.md_imported,
        RecordState.md_needs_manual_preparation,
    ]:
        raise colrev_exceptions.NotEnoughDataToIdentifyException(
            msg="cannot determine field requirements "
            "(e.g., volume/number for journal articles)",
            missing_fields=["colrev_status/field_requirements"],
        )
    # Make sure that colrev_ids are not generated when
    # identifying_field_keys are UNKNOWN but possibly required
    for identifying_field_key in FieldSet.IDENTIFYING_FIELD_KEYS:
        if record.data.get(identifying_field_key, "") == FieldValues.UNKNOWN:
            raise colrev_exceptions.NotEnoughDataToIdentifyException(
                msg=f"{identifying_field_key} unknown (maybe required)",
                missing_fields=[identifying_field_key],
            )


def _get_colrev_id_from_record(record: colrev.record.record.Record) -> str:
    try:
        # Including the version of the identifier prevents cases
        # in which almost all identifiers are identical
        # (and very few identifiers change)
        # when updating the identifier function function
        # (this may look like an anomaly and be hard to identify)
        srep = "colrev_id1:"
        if record.data[Fields.ENTRYTYPE].lower() == ENTRYTYPES.ARTICLE:
            srep = _robust_append(srep, to_append="a")
        elif record.data[Fields.ENTRYTYPE].lower() == "inproceedings":
            srep = _robust_append(srep, to_append="p")
        else:
            srep = _robust_append(
                input_string=srep, to_append=record.data[Fields.ENTRYTYPE].lower()
            )
        srep = _robust_append(
            input_string=srep,
            to_append=_get_container_title(record),
        )
        if record.data[Fields.ENTRYTYPE] == ENTRYTYPES.ARTICLE:
            # Note: volume/number may not be required.
            srep = _robust_append(
                input_string=srep, to_append=record.data.get(Fields.VOLUME, "-")
            )
            srep = _robust_append(
                input_string=srep, to_append=record.data.get(Fields.NUMBER, "-")
            )
        srep = _robust_append(srep, to_append=record.data[Fields.YEAR])
        author = _format_author_field_for_cid(record.data[Fields.AUTHOR])
        if author.replace("-", "") == "":
            raise colrev_exceptions.NotEnoughDataToIdentifyException(
                msg="Missing field:", missing_fields=[Fields.AUTHOR]
            )
        srep = _robust_append(srep, to_append=author)
        srep = _robust_append(srep, to_append=record.data[Fields.TITLE])

        srep = srep.replace(";", "")  # ";" is the separator in colrev_id list
        # Note : pages not needed.
        # pages = record_dict.get(Fields.PAGES, "")
        # srep = _robust_append(srep, pages)
    except KeyError as exc:
        if Fields.ENTRYTYPE in str(exc):
            print(f"Missing ENTRYTYPE in {record.data.get(Fields.ID, record.data)}")
        key = "unknown"
        if exc.args:
            key = exc.args[0]
        raise colrev_exceptions.NotEnoughDataToIdentifyException(
            msg="Missing field:" + str(exc), missing_fields=[key]
        )
    return srep


def get_colrev_id(record: colrev.record.record.Record, *, assume_complete: bool) -> str:
    """Create the colrev_id"""

    _check_colrev_id_preconditions(
        record,
        assume_complete=assume_complete,
    )
    srep = _get_colrev_id_from_record(record)

    # Safeguard against titles that are rarely distinct
    if any(x in srep for x in ["|minitrack-introduction"]):
        raise colrev_exceptions.NotEnoughDataToIdentifyException(
            msg="Title typically non-distinct", missing_fields=[Fields.TITLE]
        )

    return srep


def _get_colrev_pdf_id_cpid2(pdf_path: Path) -> str:
    with tempfile.NamedTemporaryFile(suffix=".png") as temp_file:
        file_name = temp_file.name
        try:
            doc: pymupdf.Document = pymupdf.open(pdf_path)
            page = next(iter(doc))  # get the first page
            pix = page.get_pixmap(dpi=200)
            pix.save(file_name)  # store image as a PNG
            with Image.open(file_name) as img:
                average_hash = imagehash.average_hash(img, hash_size=32)
                average_hash_str = str(average_hash).replace("\n", "")
                if len(average_hash_str) * "0" == average_hash_str:
                    raise colrev_exceptions.PDFHashError(path=pdf_path)
                return "cpid2:" + average_hash_str
        except StopIteration as exc:  # pragma: no cover
            raise colrev_exceptions.PDFHashError(path=pdf_path) from exc
        except pymupdf.FileDataError as exc:
            raise colrev_exceptions.InvalidPDFException(path=pdf_path) from exc
        except RuntimeError as exc:
            raise colrev_exceptions.PDFHashError(path=pdf_path) from exc


def get_colrev_pdf_id(pdf_path: Path, *, cpid_version: str = "cpid2") -> str:
    """Get the PDF hash"""

    pdf_path = pdf_path.resolve()
    if 0 == os.path.getsize(pdf_path):
        logging.error("%sPDF with size 0: %s %s", Colors.RED, pdf_path, Colors.END)
        raise colrev_exceptions.InvalidPDFException(path=pdf_path)

    if cpid_version == "cpid2":
        return _get_colrev_pdf_id_cpid2(pdf_path)

    raise NotImplementedError


def get_toc_key(record: colrev.record.record.Record) -> str:
    """Get the record's toc-key"""

    try:
        if record.data[Fields.ENTRYTYPE] == ENTRYTYPES.ARTICLE:
            toc_key = (
                record.data[Fields.JOURNAL]
                .replace(" ", "-")
                .replace("\\", "")
                .replace("&", "and")
                .lower()
            )
            toc_key += (
                f"|{record.data[Fields.VOLUME]}"
                if (
                    FieldValues.UNKNOWN
                    != record.data.get(Fields.VOLUME, FieldValues.UNKNOWN)
                )
                else "|-"
            )
            toc_key += (
                f"|{record.data[Fields.NUMBER]}"
                if (
                    FieldValues.UNKNOWN
                    != record.data.get(Fields.NUMBER, FieldValues.UNKNOWN)
                )
                else "|-"
            )

        elif record.data[Fields.ENTRYTYPE] == ENTRYTYPES.INPROCEEDINGS:
            toc_key = (
                record.data[Fields.BOOKTITLE]
                .replace(" ", "-")
                .replace("\\", "")
                .replace("&", "and")
                .lower()
                + f"|{record.data.get(Fields.YEAR, '')}"
            )
        else:
            msg = (
                f"ENTRYTYPE {record.data[Fields.ENTRYTYPE]} "
                + f"({record.data[Fields.ID]}) not toc-identifiable"
            )
            raise colrev_exceptions.NotTOCIdentifiableException(msg)
    except KeyError as exc:
        raise colrev_exceptions.NotTOCIdentifiableException(
            f"missing key {exc}"
        ) from exc

    return toc_key



#! /usr/bin/env python
"""Checker for html-tags."""
from __future__ import annotations

import re

import colrev.record.qm.quality_model
from colrev.constants import DefectCodes
from colrev.constants import Fields

# pylint: disable=too-few-public-methods


class HTMLTagChecker:
    """The HTMLTagChecker"""

    msg = DefectCodes.HTML_TAGS
    _fields_to_check = [
        Fields.TITLE,
        Fields.JOURNAL,
        Fields.BOOKTITLE,
        Fields.AUTHOR,
        Fields.PUBLISHER,
        Fields.EDITOR,
    ]

    def __init__(
        self, quality_model: colrev.record.qm.quality_model.QualityModel
    ) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.record.Record) -> None:
        """Run the html-tags checks"""

        for key in self._fields_to_check:
            if key not in record.data or record.ignored_defect(
                key=key, defect=self.msg
            ):
                continue
            if record.masterdata_is_curated():
                continue
            if re.search(r"&#\d+;", record.data[key]):
                record.add_field_provenance_note(key=key, note=self.msg)
            else:
                record.remove_field_provenance_note(key=key, note=self.msg)


def register(quality_model: colrev.record.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(HTMLTagChecker(quality_model))



#!/usr/bin/env python3
"""Repare CoLRev projects."""
from __future__ import annotations

import os
import shutil
from pathlib import Path

import colrev.env.local_index
import colrev.exceptions as colrev_exceptions
import colrev.loader.load_utils_formatter
import colrev.process.operation
from colrev.constants import DefectCodes
from colrev.constants import Fields
from colrev.constants import FieldSet
from colrev.constants import FieldValues
from colrev.constants import OperationsType
from colrev.constants import RecordState
from colrev.constants import SearchType
from colrev.writer.write_utils import write_file


class Repare(colrev.process.operation.Operation):
    """Repare a CoLRev project"""

    type = OperationsType.check

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
    ) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=self.type,
            notify_state_transition_operation=False,
        )
        # fix file no longer available
        self.local_index = colrev.env.local_index.LocalIndex(verbose_mode=False)
        self.pdf_get_operation = self.review_manager.get_pdf_get_operation(
            notify_state_transition_operation=False
        )
        self.load_formatter = colrev.loader.load_utils_formatter.LoadFormatter()

    def _fix_broken_symlink_based_on_local_index(
        self, *, record: colrev.record.record.Record, full_path: Path
    ) -> None:
        """Fix broken symlinks based on local_index"""

        if os.path.islink(full_path) and not os.path.exists(full_path):
            self.review_manager.logger.debug(f" remove broken symlink: {full_path}")
            full_path.unlink()

        try:
            retrieved_record = self.local_index.retrieve(record.data, include_file=True)

            if Fields.FILE in retrieved_record.data:
                record.update_field(
                    key=Fields.FILE,
                    value=str(retrieved_record.data[Fields.FILE]),
                    source="local_index",
                    append_edit=False,
                )
                self.pdf_get_operation.import_pdf(record)
                if Fields.FULLTEXT in retrieved_record.data:
                    del retrieved_record.data[Fields.FULLTEXT]
                self.review_manager.logger.info(
                    f" fix broken symlink: {record.data['ID']}"
                )
            else:
                self.review_manager.logger.debug(
                    f" file not linked in retrieved record: {record.data['ID']}"
                )

        except colrev_exceptions.RecordNotInIndexException:
            self.review_manager.logger.debug(
                f" record not in index: {record.data['ID']}"
            )

    def _fix_files(self, records: dict) -> None:
        # pylint: disable=too-many-branches
        for record_dict in records.values():
            if Fields.FILE not in record_dict:
                continue

            if not record_dict[Fields.FILE].startswith("data/pdfs/"):
                record_dict[Fields.FILE] = f"data/pdfs/{record_dict['ID']}.pdf"

            full_path = self.review_manager.path / Path(record_dict[Fields.FILE])

            if full_path.is_file():
                continue

            # Add .pdf extension if missing
            if Path(str(full_path) + ".pdf").is_file():
                shutil.move(str(full_path) + ".pdf", str(full_path))

            # Check / replace multiple blanks in file and filename
            try:
                parent_dir = full_path.parent
                same_dir_pdfs = [
                    x.relative_to(self.review_manager.path)
                    for x in parent_dir.glob("*.pdf")
                ]
                for same_dir_pdf in same_dir_pdfs:
                    if record_dict[Fields.FILE].replace("  ", " ") == str(
                        same_dir_pdf
                    ).replace("  ", " "):
                        shutil.move(
                            str(same_dir_pdf), str(same_dir_pdf).replace("  ", " ")
                        )
                        record_dict[Fields.FILE] = record_dict[Fields.FILE].replace(
                            "  ", " "
                        )
            except ValueError:
                pass

            full_path = self.review_manager.path / Path(record_dict[Fields.FILE])

            if not full_path.is_file():
                record = colrev.record.record.Record(record_dict)
                self._fix_broken_symlink_based_on_local_index(
                    record=record, full_path=full_path
                )

            if full_path.is_file():
                continue

            record_dict["colrev_status_backup"] = record_dict[Fields.STATUS]
            del record_dict[Fields.FILE]
            record = colrev.record.record.Record(record_dict)
            record.set_status(RecordState.rev_prescreen_included)

        file_search_sources = [
            x
            for x in self.review_manager.settings.sources
            if x.search_type == SearchType.FILES
        ]
        for search_source in file_search_sources:
            files_dir_feed = search_source.get_api_feed(
                review_manager=self.review_manager,
                source_identifier="UNKNOWN",
                update_only=True,
            )
            prefix = search_source.get_origin_prefix()
            for record_dict in records.values():
                if Fields.FILE not in record_dict:
                    continue
                for origin in record_dict[Fields.ORIGIN]:
                    if origin.startswith(prefix):
                        feed_id = origin[len(prefix) + 1 :]
                        feed_record = files_dir_feed.feed_records[feed_id]
                        if (
                            feed_record.get(Fields.FILE, "a")
                            != record_dict[Fields.FILE]
                        ):
                            feed_record[Fields.FILE] = record_dict[Fields.FILE]
                            files_dir_feed.feed_records[feed_id] = feed_record

            files_dir_feed.save()

    def _get_source_feeds(self) -> dict:
        source_feeds = {}
        for source in self.review_manager.settings.sources:
            source_feeds[str(source.filename).replace("data/search/", "")] = (
                source.get_api_feed(
                    review_manager=self.review_manager,
                    source_identifier="NA",
                    update_only=False,
                ).feed_records
            )
        return source_feeds

    # pylint: disable=too-many-branches
    def _remove_fields(self, record: colrev.record.record.Record) -> None:
        if "pdf_hash" in record.data:
            record.data[Fields.PDF_ID] = record.data["pdf_hash"]
            del record.data["pdf_hash"]
        if "pdf_prep_hints" in record.data:
            del record.data["pdf_prep_hints"]
        if Fields.GROBID_VERSION in record.data:
            del record.data[Fields.GROBID_VERSION]

        if Fields.D_PROV in record.data:
            mk_to_remove = []
            for key in record.data[Fields.D_PROV]:
                if key not in record.data:
                    mk_to_remove += [key]
            for key in mk_to_remove:
                del record.data[Fields.D_PROV][key]

        if Fields.MD_PROV in record.data:
            mdk_to_remove = []
            for key in record.data[Fields.MD_PROV]:
                if (
                    key not in record.data
                    and FieldValues.CURATED != key
                    and f"IGNORE:{DefectCodes.MISSING}"
                    not in record.data[Fields.MD_PROV][key]["note"]
                ):
                    mdk_to_remove += [key]
            for key in mdk_to_remove:
                del record.data[Fields.MD_PROV][key]

        if self.review_manager.settings.is_curated_masterdata_repo():
            if record.masterdata_is_curated():
                del record.data[Fields.MD_PROV][FieldValues.CURATED]

    def _set_data_provenance_field(
        self, *, record: colrev.record.record.Record, key: str, source_feeds: dict
    ) -> None:
        if key in record.data[Fields.D_PROV]:
            if not any(
                record.data[Fields.D_PROV][key]["source"].startswith(sf)
                for sf in list(source_feeds.keys())
            ):
                del record.data[Fields.D_PROV][key]
        if key not in record.data[Fields.D_PROV]:
            for origin in record.data[Fields.ORIGIN]:
                origin_source = origin.split("/")[0]
                origin_id = origin[len(origin_source) + 1 :]
                if (
                    origin_source not in source_feeds
                    or origin_id not in source_feeds[origin_source]
                ):
                    continue
                if key in source_feeds[origin_source][origin_id]:
                    record.add_field_provenance(key=key, source=origin, note="")
        if key == Fields.LANGUAGE:
            record.add_field_provenance(key=key, source="LanguageDetector", note="")
        if key == "tei_file":
            record.add_field_provenance(key=key, source="file|grobid", note="")
        if key == "colrev_pdf_id":
            record.add_field_provenance(key=key, source="file|pdf_hash", note="")

        if key not in record.data[Fields.D_PROV]:
            record.add_field_provenance(key=key, source="manual", note="")

        for _, prov_details in record.data[Fields.D_PROV].items():
            if prov_details["source"] in record.data[Fields.ORIGIN] + ["manual"]:
                continue
            if prov_details["source"].startswith("file|"):
                continue
            # Note : simple heuristic
            prov_details["source"] = record.data[Fields.ORIGIN][0]

    def _add_missing_masterdata_provenance(
        self,
        *,
        record: colrev.record.record.Record,
        key: str,
        value: str,
        source_feeds: dict,
    ) -> None:
        options = {}
        for origin in record.data[Fields.ORIGIN]:
            origin_source = origin.split("/")[0]
            origin_id = origin[len(origin_source) + 1 :]
            if (
                origin_source not in source_feeds
                or origin_id not in source_feeds[origin_source]
            ):
                continue
            if key in source_feeds[origin_source][origin_id]:
                options[origin_source] = source_feeds[origin_source][origin_id][key]

        source_value = "manual"

        # Note : simple heuristics:
        if value in options.values():
            if "md_curated.bib" in options:
                source_origin = "md_curated.bib"
            elif "md_crossref.bib" in options:
                source_origin = "md_crossref.bib"
            elif "CROSSREF.bib" in options:
                source_origin = "CROSSREF.bib"
            elif "md_dblp.bib" in options:
                source_origin = "md_dblp.bib"
            elif "DBLP.bib" in options:
                source_origin = "DBLP.bib"
            else:
                source_origin = [k for k, v in options.items() if v == value][0]

            for origin in record.data[Fields.ORIGIN]:
                origin_parts = origin.split("/", 1)  # Split on the first "/"
                if len(origin_parts) == 2:
                    origin_source, origin_id = origin.split("/")
                    if source_origin == origin_source:
                        source_value = origin

        record.add_field_provenance(key=key, source=source_value, note="")

    def _set_non_curated_masterdata_provenance_field(
        self,
        *,
        record: colrev.record.record.Record,
        key: str,
        value: str,
        source_feeds: dict,
    ) -> None:
        if key in record.data[Fields.MD_PROV]:
            if not any(
                record.data[Fields.MD_PROV][key]["source"].startswith(sf)
                for sf in list(source_feeds.keys())
            ):
                del record.data[Fields.MD_PROV][key]
        if key not in record.data[Fields.MD_PROV]:
            self._add_missing_masterdata_provenance(
                record=record, key=key, value=value, source_feeds=source_feeds
            )

        for prov_details in record.data[Fields.MD_PROV].values():
            if prov_details["source"] in record.data[Fields.ORIGIN] + ["manual"]:
                continue
            # Note : simple heuristic
            prov_details["source"] = record.data[Fields.ORIGIN][0]

    def _set_provenance_field(
        self,
        *,
        record: colrev.record.record.Record,
        key: str,
        value: str,
        source_feeds: dict,
    ) -> None:
        if key in FieldSet.IDENTIFYING_FIELD_KEYS:
            if record.masterdata_is_curated():
                return
            self._set_non_curated_masterdata_provenance_field(
                record=record, key=key, value=value, source_feeds=source_feeds
            )

        else:
            self._set_data_provenance_field(
                record=record, key=key, source_feeds=source_feeds
            )

    def _set_provenance(
        self, *, record: colrev.record.record.Record, source_feeds: dict
    ) -> None:
        record.align_provenance()
        for key, value in record.data.items():
            if key in [
                Fields.STATUS,
                Fields.ORIGIN,
                Fields.MD_PROV,
                Fields.D_PROV,
                "colrev_pdfid",
                "colrev_pdf_id",
                "colrev_id",
                Fields.ID,
                Fields.ENTRYTYPE,
                Fields.SCREENING_CRITERIA,
            ]:
                continue
            self._set_provenance_field(
                record=record, key=key, value=value, source_feeds=source_feeds
            )

    def _fix_provenance(self, records: dict) -> None:
        source_feeds = self._get_source_feeds()
        for record_dict in records.values():
            record = colrev.record.record.Record(record_dict)
            self._remove_fields(record)
            self._set_provenance(record=record, source_feeds=source_feeds)

    def _fix_curated_sources(self, records: dict) -> None:
        for search_source in self.review_manager.settings.sources:
            if search_source.endpoint != "colrev.local_index":
                continue

            curation_recs = colrev.loader.load_utils.load(
                filename=search_source.filename,
                logger=self.review_manager.logger,
            )

            for record_id in list(curation_recs.keys()):
                if Fields.CURATION_ID not in curation_recs[record_id]:
                    try:
                        retrieved_record = self.local_index.retrieve(
                            curation_recs[record_id], include_file=False
                        )
                        del retrieved_record.data[Fields.STATUS]
                        curation_recs[record_id] = retrieved_record.data
                    except colrev_exceptions.RecordNotInIndexException:
                        main_record_origin = (
                            search_source.get_origin_prefix() + "/" + record_id
                        )
                        main_record_l = [
                            r
                            for r in records.values()
                            if main_record_origin in r[Fields.ORIGIN]
                        ]
                        if not main_record_l:
                            continue
                        main_record_id = main_record_l[0][Fields.ID]
                        if 1 == len(records[main_record_id][Fields.ORIGIN]):
                            del records[main_record_id]
                        else:
                            records[main_record_id][Fields.ORIGIN].remove(
                                main_record_origin
                            )
                        del curation_recs[record_id]

            write_file(records_dict=curation_recs, filename=search_source.filename)
            self.review_manager.dataset.add_changes(search_source.filename)

    def _update_field_names(self, records: dict) -> None:
        for record_dict in records.values():
            # TBD: which parts are in upgrade/repare and which parts are in prepare??
            record = colrev.record.record.Record(record_dict)
            if Fields.FULLTEXT in record_dict.get("link", ""):
                record.rename_field(key="link", new_key=Fields.FULLTEXT)
            if (
                record.data.get("note", "").lower().startswith("cited by ")
                and Fields.CITED_BY not in record.data
            ):
                record.data["note"] = record.data["note"][9:]
                record.rename_field(key="note", new_key=Fields.CITED_BY)
            if (
                record.data.get("note", "").lower().startswith("cited by ")
                and Fields.CITED_BY in record.data
            ):
                record.remove_field(key="note")
            if record.data.get("link", "").startswith(
                "https://api.elsevier.com/content/article/"
            ) and record.data.get("link", "").endswith("Accept=text/xml"):
                record.remove_field(key="link")

    def _fix_field_values(self, records: dict) -> None:
        for record_dict in records.values():
            record = colrev.record.record.Record(record_dict)
            self.load_formatter.run(record)

    @colrev.process.operation.Operation.decorate()
    def main(self) -> None:
        """Repare a CoLRev project (main entrypoint)"""

        # Try: open settings, except: notify & start repare

        # ...

        # Try: open records, except: notify & start repare
        try:
            records = self.review_manager.dataset.load_records_dict()
        except AttributeError:
            self.review_manager.logger.error("Could not read bibtex file")

            separated_records = {}  # type: ignore
            with open(self.review_manager.paths.records, encoding="utf-8") as file:
                record_str = ""
                line = file.readline()

                while line:
                    if line == "\n":
                        records = colrev.loader.load_utils.loads(
                            load_string=record_str,
                            implementation="bib",
                            logger=self.review_manager.logger,
                        )

                        if len(records) != 1:
                            print(record_str)
                        else:
                            separated_records = {**separated_records, **records}
                        record_str = ""
                    record_str += line
                    line = file.readline()

            write_file(records_dict=separated_records, filename=Path("extracted.bib"))

            try:
                records = self.review_manager.dataset.load_records_dict()
            except AttributeError:
                return

        self._fix_curated_sources(records)

        # removing specific fields
        # for record_dict in records.values():
        #     if "colrev_status_backup" in record_dict:
        #         colrev.record.record.Record(record_dict).remove_field(
        #             key="colrev_status_backup"
        #         )
        #     if "colrev_local_index" in record_dict:
        #         colrev.record.record.Record(record_dict).remove_field(
        #             key="colrev_local_index"
        #         )

        self._update_field_names(records)

        self._fix_provenance(records)

        self._fix_field_values(records)

        self._fix_files(records)

        self.review_manager.dataset.save_records_dict(records)



#!/usr/bin/env python3
"""Hook for sharing in CoLRev repositories"""
import colrev.review_manager


def main() -> int:
    """Main entrypoint for the hook"""

    review_manager = colrev.review_manager.ReviewManager()
    ret = review_manager.sharing()

    print(ret["msg"])

    return ret["status"]


if __name__ == "__main__":
    raise SystemExit(main())



#! /usr/bin/env python
"""Checker for identical-values-between-title-and-container."""
from __future__ import annotations

import colrev.record.qm.quality_model
from colrev.constants import DefectCodes
from colrev.constants import Fields
from colrev.constants import FieldValues

# pylint: disable=too-few-public-methods


class IdenticalValuesChecker:
    """The IdenticalValuesChecker"""

    msg = DefectCodes.IDENTICAL_VALUES_BETWEEN_TITLE_AND_CONTAINER

    def __init__(
        self, quality_model: colrev.record.qm.quality_model.QualityModel
    ) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.record.Record) -> None:
        """Run the identical-values-between-title-and-container checks"""

        if record.ignored_defect(key=Fields.TITLE, defect=self.msg):
            return

        if self._identical_values_between_title_and_container(record=record):
            record.add_field_provenance_note(key=Fields.TITLE, note=self.msg)
        else:
            record.remove_field_provenance_note(key=Fields.TITLE, note=self.msg)

    def _identical_values_between_title_and_container(
        self, *, record: colrev.record.record.Record
    ) -> bool:
        if record.data.get(Fields.TITLE, FieldValues.UNKNOWN) == FieldValues.UNKNOWN:
            return False
        if (
            Fields.BOOKTITLE in record.data
            and Fields.TITLE in record.data
            and record.data[Fields.TITLE].lower().replace("the ", "")
            == record.data[Fields.BOOKTITLE].lower().replace("the ", "")
        ):
            return True
        if (
            Fields.JOURNAL in record.data
            and Fields.TITLE in record.data
            and record.data[Fields.TITLE].lower().replace("the ", "")
            == record.data[Fields.JOURNAL].lower().replace("the ", "")
        ):
            return True
        return False


def register(quality_model: colrev.record.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(IdenticalValuesChecker(quality_model))



#! /usr/bin/env python
"""Preparation of curations"""
from __future__ import annotations

import zope.interface
from pydantic import Field

import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields
from colrev.constants import RecordState


# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.PrepInterface)
class CurationPrep:
    """Preparation of curations"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings

    ci_supported: bool = Field(default=True)

    source_correction_hint = "check with the developer"
    always_apply_changes = True

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,
        settings: dict,
    ) -> None:
        self.settings = self.settings_class(**settings)
        self.quality_model = prep_operation.review_manager.get_qm()
        self.prep_operation = prep_operation
        self.review_manager = prep_operation.review_manager
        self.curation_restrictions = self._load_curation_restrictions()

    def _load_curation_restrictions(self) -> dict:
        curation_restrictions = {}
        curated_endpoints = [
            x
            for x in self.review_manager.settings.data.data_package_endpoints
            if x["endpoint"] == "colrev.colrev_curation"
        ]
        if curated_endpoints:
            curated_endpoint = curated_endpoints[0]
            curation_restrictions = curated_endpoint.get("masterdata_restrictions", {})
        return curation_restrictions

    def _get_applicable_curation_restrictions(
        self, *, record: colrev.record.record.Record
    ) -> dict:
        """Get the applicable curation restrictions"""

        if not str(record.data.get(Fields.YEAR, "NA")).isdigit():
            return {}

        start_year_values = list(self.curation_restrictions.keys())
        year_index_diffs = [
            int(record.data[Fields.YEAR]) - int(x) for x in start_year_values
        ]
        year_index_diffs = [x if x >= 0 else 2000 for x in year_index_diffs]

        if not year_index_diffs:
            return {}

        index_min = min(range(len(year_index_diffs)), key=year_index_diffs.__getitem__)
        applicable_curation_restrictions = self.curation_restrictions[
            start_year_values[index_min]
        ]
        return applicable_curation_restrictions

    def apply_curation_restrictions(
        self, *, record: colrev.record.record.Record
    ) -> None:
        """Apply the curation restrictions"""
        applicable_curation_restrictions = self._get_applicable_curation_restrictions(
            record=record
        )
        if Fields.ENTRYTYPE in applicable_curation_restrictions:
            if applicable_curation_restrictions[Fields.ENTRYTYPE] != record.data.get(
                Fields.ENTRYTYPE, ""
            ):
                try:
                    record.change_entrytype(
                        new_entrytype=applicable_curation_restrictions[
                            Fields.ENTRYTYPE
                        ],
                        qm=self.quality_model,
                    )
                except colrev_exceptions.MissingRecordQualityRuleSpecification as exc:
                    print(exc)

        for field in [Fields.JOURNAL, Fields.BOOKTITLE]:
            if field not in applicable_curation_restrictions:
                continue
            if applicable_curation_restrictions[field] == record.data.get(field, ""):
                continue
            record.update_field(
                key=field,
                value=applicable_curation_restrictions[field],
                source="colrev_curation.curation_restrictions",
            )

        if (
            Fields.VOLUME in applicable_curation_restrictions
            and not applicable_curation_restrictions[Fields.VOLUME]
            and Fields.VOLUME in record.data
        ):
            record.remove_field(
                key=Fields.VOLUME,
                not_missing_note=True,
                source="colrev_curation.curation_restrictions",
            )

        if (
            Fields.NUMBER in applicable_curation_restrictions
            and not applicable_curation_restrictions[Fields.NUMBER]
            and Fields.NUMBER in record.data
        ):
            record.remove_field(
                key=Fields.NUMBER,
                not_missing_note=True,
                source="colrev_curation.curation_restrictions",
            )

    def prepare(
        self,
        record: colrev.record.record_prep.PrepRecord,
    ) -> colrev.record.record.Record:
        """Prepare records in a CoLRev curation"""

        if record.data[Fields.STATUS] == RecordState.rev_prescreen_excluded:
            return record

        self.apply_curation_restrictions(record=record)

        return record



#! /usr/bin/env python
"""Screen based on a table"""
from __future__ import annotations

import csv
import typing
from pathlib import Path

import pandas as pd
import zope.interface
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.packages.screen_utils as util_cli_screen
import colrev.record.record
import colrev.settings
from colrev.constants import Fields
from colrev.constants import RecordState


@zope.interface.implementer(colrev.package_manager.interfaces.ScreenInterface)
class TableScreen:
    """Screen documents using tables (exported and imported)"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings
    ci_supported: bool = Field(default=False)
    export_todos_only: bool = True

    screen_table_path = Path("screen/screen.csv")

    def __init__(
        self,
        *,
        screen_operation: colrev.ops.screen.Screen,
        settings: dict,
    ) -> None:
        self.review_manager = screen_operation.review_manager
        self.screen_operation = screen_operation
        self.settings = self.settings_class(**settings)

    def _create_screening_table(self, *, records: dict, split: list) -> list:
        # pylint: disable=too-many-branches
        self.review_manager.logger.info("Loading records for export")

        screening_criteria = util_cli_screen.get_screening_criteria_from_user_input(
            screen_operation=self.screen_operation, records=records
        )

        tbl = []
        for record in records.values():
            if record[Fields.STATUS] not in [
                RecordState.pdf_prepared,
            ]:
                continue

            if len(split) > 0:
                if record[Fields.ID] not in split:
                    continue

            inclusion_2 = "NA"

            if RecordState.pdf_prepared == record[Fields.STATUS]:
                inclusion_2 = "TODO"
            elif self.export_todos_only:
                continue
            if RecordState.rev_excluded == record[Fields.STATUS]:
                inclusion_2 = "out"
            if record[Fields.STATUS] in [
                RecordState.rev_included,
                RecordState.rev_synthesized,
            ]:
                inclusion_2 = "in"

            # pylint: disable=duplicate-code
            row = {
                Fields.ID: record[Fields.ID],
                Fields.AUTHOR: record.get(Fields.AUTHOR, ""),
                Fields.TITLE: record.get(Fields.TITLE, ""),
                Fields.JOURNAL: record.get(Fields.JOURNAL, ""),
                Fields.BOOKTITLE: record.get(Fields.BOOKTITLE, ""),
                Fields.YEAR: record.get(Fields.YEAR, ""),
                Fields.VOLUME: record.get(Fields.VOLUME, ""),
                Fields.NUMBER: record.get(Fields.NUMBER, ""),
                Fields.PAGES: record.get(Fields.PAGES, ""),
                Fields.DOI: record.get(Fields.DOI, ""),
                Fields.ABSTRACT: record.get(Fields.ABSTRACT, ""),
            }

            if len(screening_criteria) == 0:
                # No criteria: code inclusion directly
                row["screen_inclusion"] = inclusion_2

            else:
                # Code criteria
                screening_criteria_field = record.get(Fields.SCREENING_CRITERIA, "")
                if screening_criteria_field == "":
                    # and inclusion_2 == "in"
                    for criterion_name in screening_criteria.keys():
                        row[criterion_name] = "TODO (in/out)"
                else:
                    for criterion_name, decision in screening_criteria_field.split(";"):
                        row[criterion_name] = decision

            tbl.append(row)

        return tbl

    def export_table(
        self,
        records: dict,
        split: list,
        export_table_format: str = "csv",
    ) -> None:
        """Export a screening table"""

        # gh_issue https://github.com/CoLRev-Environment/colrev/issues/73
        # add delta (records not yet in the table)
        # instead of overwriting
        # export_table_format as a settings parameter

        if self.screen_table_path.is_file():
            print("File already exists. Please rename it.")
            return

        tbl = self._create_screening_table(records=records, split=split)

        self.screen_table_path.parents[0].mkdir(parents=True, exist_ok=True)

        if export_table_format.lower() == "csv":
            screen_df = pd.DataFrame(tbl)
            screen_df.to_csv(self.screen_table_path, index=False, quoting=csv.QUOTE_ALL)
            self.review_manager.logger.info(f"Created {self.screen_table_path}")

        if export_table_format.lower() == "xlsx":
            screen_df = pd.DataFrame(tbl)
            screen_df.to_excel(
                self.screen_table_path.with_suffix(".xlsx"),
                index=False,
                sheet_name="screen",
            )
            self.review_manager.logger.info(
                f"Created {self.screen_table_path.with_suffix('.xlsx')}"
            )

        return

    def import_table(
        self,
        records: dict,
        import_table_path: typing.Optional[Path] = None,
    ) -> None:
        """Import a screening table"""

        # pylint: disable=duplicate-code
        if import_table_path is None:
            import_table_path = self.screen_table_path

        if not Path(import_table_path).is_file():
            self.review_manager.logger.error(
                f"Did not find {import_table_path} - exiting."
            )
            return

        screen_df = pd.read_csv(import_table_path)
        screen_df.fillna("", inplace=True)
        screened_records = screen_df.to_dict("records")

        screening_criteria = self.review_manager.settings.screen.criteria

        for screened_record in screened_records:
            if screened_record.get(Fields.ID, "") in records:
                record_dict = records[screened_record.get(Fields.ID, "")]
                record = colrev.record.record.Record(record_dict)
                if "screen_inclusion" in screened_record:
                    if screened_record["screen_inclusion"] == "in":
                        record.set_status(RecordState.rev_included)
                    elif screened_record["screen_inclusion"] == "out":
                        record.set_status(RecordState.rev_excluded)
                    else:
                        print(
                            f"Invalid choice: {screened_record['screen_inclusion']} "
                            f"({screened_record['ID']})"
                        )
                    continue
                screening_criteria_field = ""
                for screening_criterion in screening_criteria.keys():
                    assert screened_record[screening_criterion] in ["in", "out"]
                    screening_criteria_field += (
                        screening_criterion
                        + "="
                        + screened_record[screening_criterion]
                        + ";"
                    )
                screening_criteria_field = screening_criteria_field.rstrip(";")
                record.data[Fields.SCREENING_CRITERIA] = screening_criteria_field
                if "=out" in screening_criteria_field:
                    record.set_status(RecordState.rev_excluded)
                else:
                    record.set_status(RecordState.rev_included)

        self.review_manager.dataset.save_records_dict(records)

    def run_screen(self, records: dict, split: list) -> dict:
        """Screen records based on screening tables"""

        if input("create screen table [y,n]?") == "y":
            self.export_table(records, split)

        if input("import screen table [y,n]?") == "y":
            self.import_table(records)

        if self.review_manager.dataset.has_record_changes():
            if input("create commit [y,n]?") == "y":
                self.review_manager.dataset.create_commit(
                    msg="Screen", manual_author=True
                )
        return records



#! /usr/bin/env python
"""CoLRev prep operation: Prepare record metadata."""
from __future__ import annotations

import inspect
import logging
import multiprocessing as mp
import random
import shutil
import typing
from copy import deepcopy
from datetime import datetime
from datetime import timedelta
from multiprocessing import Lock
from multiprocessing import Value
from multiprocessing.pool import ThreadPool as Pool
from pathlib import Path

from requests.exceptions import ConnectionError as requests_ConnectionError
from requests.exceptions import ReadTimeout

import colrev.env.utils
import colrev.exceptions as colrev_exceptions
import colrev.loader.load_utils
import colrev.process.operation
import colrev.record.record_prep
from colrev.constants import Colors
from colrev.constants import DefectCodes
from colrev.constants import EndpointType
from colrev.constants import Fields
from colrev.constants import FieldSet
from colrev.constants import OperationsType
from colrev.constants import RecordState
from colrev.writer.write_utils import to_string
from colrev.writer.write_utils import write_file

if typing.TYPE_CHECKING:  # pragma: no cover
    import colrev.package_manager.interfaces
    import colrev.review_manager
    import colrev.settings

# logging.getLogger("urllib3").setLevel(logging.ERROR)
logging.getLogger("requests_cache").setLevel(logging.ERROR)

PREP_COUNTER = Value("i", 0)


class PreparationBreak(Exception):
    """Event interrupting the preparation."""


# pylint: disable=duplicate-code
FIELDS_TO_KEEP = FieldSet.STANDARDIZED_FIELD_KEYS + [
    Fields.DBLP_KEY,
    Fields.SEMANTIC_SCHOLAR_ID,
    Fields.WEB_OF_SCIENCE_ID,
    Fields.EDITION,
]


# pylint: disable=too-many-instance-attributes
class Prep(colrev.process.operation.Operation):
    """Prepare records (metadata)"""

    timeout = 30
    max_retries_on_error = 3
    pad: int = 0

    first_round: bool
    last_round: bool

    polish: bool = False

    prep_package_endpoints: dict[str, typing.Any]

    _cpu = 1
    _prep_commit_id = "HEAD"

    type = OperationsType.prep

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        notify_state_transition_operation: bool,
        polish: bool,
        cpu: int,
    ) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=self.type,
            notify_state_transition_operation=notify_state_transition_operation,
        )
        self.fields_to_keep = (
            FIELDS_TO_KEEP + self.review_manager.settings.prep.fields_to_keep
        )

        self._stats: typing.Dict[str, typing.List[timedelta]] = {}

        self.temp_prep_lock = Lock()
        self.current_temp_records = self.review_manager.path / Path(
            ".colrev/cur_temp_recs.bib"
        )

        self.temp_records = self.review_manager.path / (Path(".colrev/temp_recs.bib"))

        self.quality_model = review_manager.get_qm()
        self.package_manager = self.review_manager.get_package_manager()

        self.polish = polish
        self._cpu = cpu

        # Note: for unit testing, we use a simple loop (instead of parallel)
        # to ensure that the IDs of feed records don't change
        unit_testing = "test_prep" == inspect.stack()[1][3]
        if unit_testing:
            self._cpu = 1

    def _add_stats(
        self, *, prep_round_package_endpoint: dict, start_time: datetime
    ) -> None:
        if prep_round_package_endpoint["endpoint"] not in self._stats:
            self._stats[prep_round_package_endpoint["endpoint"]] = [
                datetime.now() - start_time
            ]
        else:
            self._stats[prep_round_package_endpoint["endpoint"]].append(
                datetime.now() - start_time
            )

    def _print_stats(self) -> None:
        if self.review_manager.verbose_mode:
            print("Runtime statistics (averages)")
            averaged_list = [
                {
                    "script": script,
                    "average": sum(deltalist, timedelta(0)) / len(deltalist),
                }
                for script, deltalist in self._stats.items()
            ]
            for item in sorted(
                averaged_list,
                key=lambda k: k["average"],  # type: ignore
                reverse=True,
            ):
                average_time_str = (
                    f"{item['average'].seconds}."  # type: ignore
                    f"{item['average'].microseconds}"  # type: ignore
                )
                average_time = float(average_time_str)
                average_time = round(average_time, 2)
                average_time_str = f"{average_time:.2f}"
                print(
                    f"{item['script']} ".ljust(50, " ")
                    + ":"
                    + f"{average_time_str} s".rjust(10, " ")
                )
            print()

    def _print_diffs_for_debug(
        self,
        *,
        prior: colrev.record.record_prep.PrepRecord,
        preparation_record: colrev.record.record_prep.PrepRecord,
        prep_package_endpoint: colrev.package_manager.interfaces.PrepInterface,
    ) -> None:
        pass  # this method can be replaced by inheriting class (for debugging)

    def _package_prep(
        self,
        prep_round_package_endpoint: dict,
        record: colrev.record.record_prep.PrepRecord,
        preparation_record: colrev.record.record_prep.PrepRecord,
    ) -> None:

        try:
            if (
                prep_round_package_endpoint["endpoint"].lower()
                not in self.prep_package_endpoints
            ):
                return
            endpoint = self.prep_package_endpoints[
                prep_round_package_endpoint["endpoint"].lower()
            ]

            prior = preparation_record.copy_prep_rec()

            start_time = datetime.now()
            preparation_record = endpoint.prepare(preparation_record)
            self._add_stats(
                start_time=start_time,
                prep_round_package_endpoint=prep_round_package_endpoint,
            )

            self._print_diffs_for_debug(
                prior=prior,
                preparation_record=preparation_record,
                prep_package_endpoint=endpoint,
            )

            if endpoint.always_apply_changes:
                record.update_by_record(preparation_record)

            if self._preparation_save_condition(preparation_record):
                record.update_by_record(preparation_record)

            if (
                self._preparation_break_condition(preparation_record)
                and not self.polish
            ):
                record.update_by_record(preparation_record)
                raise PreparationBreak
        except ReadTimeout:
            self._add_stats(
                start_time=start_time,
                prep_round_package_endpoint=prep_round_package_endpoint,
            )
            if self.review_manager.verbose_mode:
                self.review_manager.logger.error(
                    f" {Colors.RED}{record.data[Fields.ID]}".ljust(45)
                    + f"{endpoint.settings.endpoint}(...) timed out{Colors.END}{Colors.END}"
                )

        except colrev_exceptions.ServiceNotAvailableException as exc:
            if self.review_manager.force_mode:
                self._add_stats(
                    start_time=start_time,
                    prep_round_package_endpoint=prep_round_package_endpoint,
                )
                self.review_manager.logger.error(exc)
            else:
                raise exc

    def _print_post_package_prep_polish_info(
        self,
        *,
        record: colrev.record.record_prep.PrepRecord,
        prior_state: RecordState,
        progress: str,
    ) -> None:
        # records in post_md_prepared remain in that state (in polish mode)
        if (
            record.data[Fields.STATUS]
            in RecordState.get_post_x_states(state=RecordState.md_prepared)
            and prior_state != RecordState.md_needs_manual_preparation
        ) or (
            prior_state == RecordState.md_needs_manual_preparation
            and record.data[Fields.STATUS] == RecordState.md_needs_manual_preparation
        ):
            self.review_manager.logger.info(
                f" {record.data[Fields.ID]}".ljust(41) + f"{progress} - "
            )
        elif record.data[Fields.STATUS] == RecordState.rev_prescreen_excluded:
            self.review_manager.logger.info(
                f" {Colors.RED}{record.data[Fields.ID]}".ljust(46)
                + f"{progress}{prior_state} →  {record.data[Fields.STATUS]}"
                + f"{Colors.END}"
                + f" ({record.data.get('prescreen_exclusion', 'NA')})"
            )
        elif record.data[Fields.STATUS] == RecordState.md_needs_manual_preparation:
            self.review_manager.logger.info(
                f" {Colors.ORANGE}{record.data[Fields.ID]}".ljust(46)
                + f"{progress}{prior_state} →  {record.data[Fields.STATUS]}{Colors.END}"
            )
        elif record.data[Fields.STATUS] == RecordState.md_prepared:
            curation_addition = "   "
            if record.masterdata_is_curated():
                curation_addition = " ✔ "
            self.review_manager.logger.info(
                f" {Colors.GREEN}{record.data[Fields.ID]}".ljust(46)
                + f"{progress}{prior_state} →  "
                f"{record.data[Fields.STATUS]}{Colors.END}{curation_addition}"
            )

        elif record.data[Fields.STATUS] == RecordState.md_needs_manual_preparation:
            self.review_manager.logger.info(
                f" {Colors.ORANGE}{record.data[Fields.ID]}".ljust(46)
                + f"{progress}{prior_state} →  {record.data[Fields.STATUS]}{Colors.END}"
            )

    def _preparation_break_condition(
        self, record: colrev.record.record_prep.PrepRecord
    ) -> bool:
        """Check whether the break condition for the prep operation is given"""

        if DefectCodes.RECORD_NOT_IN_TOC in record.get_field_provenance_notes(
            Fields.JOURNAL
        ):
            return True
        if DefectCodes.RECORD_NOT_IN_TOC in record.get_field_provenance_notes(
            Fields.BOOKTITLE
        ):
            return True

        if record.data.get(Fields.STATUS, "NA") in [
            RecordState.rev_prescreen_excluded,
        ]:
            return True
        return False

    def _preparation_save_condition(
        self, record: colrev.record.record_prep.PrepRecord
    ) -> bool:
        """Check whether the save condition for the prep operation is given"""

        if record.data[Fields.STATUS] in [
            RecordState.rev_prescreen_excluded,
            RecordState.md_prepared,
        ]:
            return True

        if DefectCodes.RECORD_NOT_IN_TOC in record.get_field_provenance_notes(
            Fields.JOURNAL
        ):
            return True
        if DefectCodes.RECORD_NOT_IN_TOC in record.get_field_provenance_notes(
            Fields.BOOKTITLE
        ):
            return True

        return False

    def _status_to_prepare(self, record: colrev.record.record_prep.PrepRecord) -> bool:
        """Check whether the record needs to be prepared"""
        return record.data.get(Fields.STATUS, "NA") in [
            RecordState.md_needs_manual_preparation,
            RecordState.md_imported,
            RecordState.md_prepared,
        ]

    def _print_post_package_prep_info(
        self,
        record: colrev.record.record_prep.PrepRecord,
        item: dict,
        prior_state: RecordState,
    ) -> None:
        # pylint: disable=redefined-outer-name,invalid-name
        with PREP_COUNTER.get_lock():
            PREP_COUNTER.value += 1  # type: ignore
        progress = ""
        if item["nr_items"] > 100:
            progress = f"({PREP_COUNTER.value}/{item['nr_items']}) ".rjust(  # type: ignore
                12, " "
            )

        if self.polish:
            self._print_post_package_prep_polish_info(
                record=record, prior_state=prior_state, progress=progress
            )
            return

        if self._preparation_break_condition(record):
            if RecordState.rev_prescreen_excluded == record.data[Fields.STATUS]:
                target_state = RecordState.rev_prescreen_excluded
                self.review_manager.logger.info(
                    f" {Colors.RED}{record.data[Fields.ID]}".ljust(46)
                    + f"{progress}{prior_state} →  {target_state}"
                    + f"{Colors.END}"
                    + f" ({record.data.get('prescreen_exclusion', 'NA')})"
                )
            else:
                target_state = RecordState.md_needs_manual_preparation
                self.review_manager.logger.info(
                    f" {Colors.ORANGE}{record.data[Fields.ID]}".ljust(46)
                    + f"{progress}{prior_state} →  {target_state}{Colors.END}"
                )

        elif self._preparation_save_condition(record):
            curation_addition = "   "
            if record.masterdata_is_curated():
                curation_addition = " ✔ "
            target_state = RecordState.md_prepared
            self.review_manager.logger.info(
                f" {Colors.GREEN}{record.data[Fields.ID]}".ljust(46)
                + f"{progress}{prior_state} →  {target_state}{Colors.END}{curation_addition}"
            )
        else:
            target_state = RecordState.md_needs_manual_preparation
            self.review_manager.logger.info(
                f" {Colors.ORANGE}{record.data[Fields.ID]}".ljust(46)
                + f"{progress}{prior_state} →  {target_state}{Colors.END}"
            )

    def _post_package_prep(
        self,
        record: colrev.record.record_prep.PrepRecord,
        preparation_record: colrev.record.record_prep.PrepRecord,
        item: dict,
        prior_state: RecordState,
    ) -> None:
        if self.polish:
            record.data[Fields.STATUS] = prior_state

        if self.last_round and not self.polish:
            if self._status_to_prepare(record):
                for key in list(record.data.keys()):
                    if key not in self.fields_to_keep:
                        record.remove_field(key=key)
                    elif record.data[key] in ["", "NA"]:
                        record.remove_field(key=key)
                record.update_by_record(preparation_record)

        # Note: run_quality_model sets to md_needs_manual_preparation
        record.run_quality_model(self.quality_model, set_prepared=not self.polish)

        if not self.review_manager.verbose_mode:
            self._print_post_package_prep_info(
                record=record, item=item, prior_state=prior_state
            )

    def _save_to_temp(self, record: colrev.record.record_prep.PrepRecord) -> None:
        rec_str = to_string(
            records_dict={record.data[Fields.ID]: record.get_data()},
            implementation="bib",
        )
        self.temp_prep_lock.acquire(timeout=120)
        self.current_temp_records.parent.mkdir(exist_ok=True)
        with open(self.current_temp_records, "a", encoding="utf-8") as cur_temp_rec:
            cur_temp_rec.write(rec_str)
        try:
            self.temp_prep_lock.release()
        except ValueError:
            pass

    def _complete_resumed_operation(self, prepared_records: list) -> None:
        if self.temp_records.is_file():
            temp_recs = colrev.loader.load_utils.load(
                filename=self.temp_records,
                logger=self.review_manager.logger,
            )
            prepared_records_ids = [x[Fields.ID] for x in prepared_records]
            for record in temp_recs.values():
                if record[Fields.ID] not in prepared_records_ids:
                    prepared_records.append(record)

        self.temp_records.unlink(missing_ok=True)
        self.current_temp_records.unlink(missing_ok=True)

    def _validate_record(
        self,
        *,
        record: colrev.record.record_prep.PrepRecord,
        prep_round_package_endpoint: str,
    ) -> None:
        if Fields.STATUS not in record.data:
            print(record.data)
            raise ValueError(
                f"Record {record.data[Fields.ID]} has no colrev_status"
                f" after {prep_round_package_endpoint}"
            )
        if not self.polish and record.data[Fields.STATUS] not in [
            RecordState.md_imported,
            RecordState.md_prepared,
            RecordState.md_needs_manual_preparation,
            RecordState.rev_prescreen_excluded,
        ]:
            print(record.data)
            raise ValueError(
                f"Record {record.data[Fields.ID]} has invalid status {record.data[Fields.STATUS]}"
                f" after {prep_round_package_endpoint}"
            )
        if Fields.MD_PROV not in record.data:
            raise ValueError(
                f"Record {record.data[Fields.ID]} has no Fields.MD_PROV"
                f" after {prep_round_package_endpoint}"
            )
        if Fields.ID not in record.data:
            raise ValueError(
                f"Record {record.data[Fields.ID]} has no ID"
                f" after {prep_round_package_endpoint}"
            )
        if Fields.ENTRYTYPE not in record.data:
            raise ValueError(
                f"Record {record.data[Fields.ID]} has no ENTRYTYPE"
                f" after {prep_round_package_endpoint}"
            )

    # Note : no named arguments for multiprocessing
    def prepare(self, item: dict) -> dict:
        """Prepare a record (based on package_endpoints in the settings)"""

        # https://docs.python.org/3/library/concurrent.futures.html
        # #concurrent.futures.Executor.map
        # Exceptions are raised at the end/when results are retrieved from the iterator

        record: colrev.record.record_prep.PrepRecord = item["record"]

        if not self._status_to_prepare(record) and not self.polish:
            return record.get_data()

        if self.review_manager.verbose_mode:
            self.review_manager.logger.info(" prep " + record.data[Fields.ID])

        record.require_prov()

        # preparation_record changes with each endpoint and
        # eventually replaces record (if md_prepared or endpoint.always_apply_changes)
        preparation_record = record.copy_prep_rec()
        prior_state = record.data[Fields.STATUS]

        # Rerun quality model (in case there are manual prep changes)
        preparation_record.change_entrytype(
            new_entrytype=record.data[Fields.ENTRYTYPE], qm=self.quality_model
        )
        preparation_record.run_quality_model(
            self.quality_model, set_prepared=not self.polish
        )

        for prep_round_package_endpoint in deepcopy(
            item["prep_round_package_endpoints"]
        ):
            try:
                self._package_prep(
                    prep_round_package_endpoint,
                    record,
                    preparation_record,
                )
                self._validate_record(
                    record=record,
                    prep_round_package_endpoint=prep_round_package_endpoint,
                )
                # Note: ServiceNotAvailableException should be ignored
                # in the packages if review_manager.force_mode
            except PreparationBreak:
                break

        self._post_package_prep(
            record=record,
            preparation_record=preparation_record,
            item=item,
            prior_state=prior_state,
        )

        self._save_to_temp(record)

        return record.get_data()

    def _rename_files(self, records: dict) -> None:
        def file_rename_condition(record_dict: dict) -> bool:
            if Fields.FILE not in record_dict:
                return False
            if not Path(record_dict[Fields.FILE]).is_file():
                return False
            if (
                str(Path(record_dict[Fields.FILE]).name)
                == f"{record_dict[Fields.ID]}.pdf"
            ):
                return False

            return True

        for record_dict in records.values():
            if not file_rename_condition(record_dict):
                continue

            old_filename = record_dict[Fields.FILE]
            new_filename = Path(record_dict[Fields.FILE]).parent / Path(
                f"{record_dict[Fields.ID]}.pdf"
            )
            shutil.move(record_dict[Fields.FILE], str(new_filename))
            record_dict[Fields.FILE] = str(new_filename)

            # simple heuristic:
            pdfs_origin_file = Path("data/search/pdfs.bib")
            if pdfs_origin_file.is_file():
                colrev.env.utils.inplace_change(
                    filename=pdfs_origin_file,
                    old_string=old_filename,
                    new_string=str(new_filename),
                )
                self.review_manager.dataset.add_changes(pdfs_origin_file)

    def set_ids(self) -> None:
        """Set IDs (regenerate). In force-mode, all IDs are regenerated and PDFs are renamed"""

        self.review_manager.logger.info("Set IDs")
        records = self.review_manager.dataset.load_records_dict()
        records = self.review_manager.dataset.set_ids()
        self._rename_files(records)
        self.review_manager.dataset.save_records_dict(records)
        self.review_manager.dataset.create_commit(msg="Set IDs")

    def setup_custom_script(self) -> None:
        """Setup a custom prep script"""

        filedata = colrev.env.utils.get_package_file_content(
            module="colrev.ops", filename=Path("custom_scripts/custom_prep_script.py")
        )
        if filedata:
            with open("custom_prep_script.py", "w", encoding="utf-8") as file:
                file.write(filedata.decode("utf-8"))

        self.review_manager.dataset.add_changes(Path("custom_prep_script.py"))

        prep_round = self.review_manager.settings.prep.prep_rounds[-1]
        prep_round.prep_package_endpoints.append({"endpoint": "custom_prep_script"})
        self.review_manager.save_settings()

    def _load_prep_data(self) -> dict:
        records_headers = self.review_manager.dataset.load_records_dict(
            header_only=True
        )
        record_header_list = list(records_headers.values())

        pad = (
            35
            if (0 == len(record_header_list))
            else min((max(len(x[Fields.ID]) for x in record_header_list) + 2), 35)
        )

        r_states_to_prepare = [
            RecordState.md_imported,
            RecordState.md_needs_manual_preparation,
        ]
        if self.polish:
            r_states_to_prepare = list(RecordState)

        items = list(
            self.review_manager.dataset.read_next_record(
                conditions=[{Fields.STATUS: s} for s in r_states_to_prepare]
            )
        )
        if (
            self.polish
            and self.review_manager.in_ci_environment()
            and len(items) > 2000
        ):
            items = random.choices(items, k=2000)  # nosec

        prep_data = {
            "nr_tasks": len(items),
            "PAD": pad,
            "items": list(items),
        }

        return prep_data

    def _load_temp_prep_to_resume(self, prepare_data: dict) -> None:
        if self.current_temp_records.is_file():
            # combine and remove redundant records

            cur_temp_recs = colrev.loader.load_utils.load(
                filename=self.current_temp_records,
                logger=self.review_manager.logger,
            )

            temp_recs = {}
            if self.temp_records.is_file():
                temp_recs = colrev.loader.load_utils.load(
                    filename=self.temp_records,
                    logger=self.review_manager.logger,
                )

            combined_recs = {**temp_recs, **cur_temp_recs}
            self.temp_records.parent.mkdir(exist_ok=True)

            write_file(records_dict=combined_recs, filename=self.temp_records)

            self.current_temp_records.unlink()

        if self.temp_records.is_file():

            temp_recs = colrev.loader.load_utils.load(
                filename=self.temp_records,
                logger=self.review_manager.logger,
            )

            self.review_manager.logger.info("Continue with existing records")
            skipped_items = 0
            list_to_skip = []
            for item in prepare_data["items"]:
                if item[Fields.ID] not in temp_recs:
                    continue
                del temp_recs[item[Fields.ID]]
                list_to_skip.append(item[Fields.ID])
                skipped_items += 1
            self.review_manager.logger.info(
                f"{Colors.GREEN}Skipped {skipped_items} records{Colors.END}"
            )
            prepare_data["items"] = [
                x for x in prepare_data["items"] if x[Fields.ID] not in list_to_skip
            ]

            with PREP_COUNTER.get_lock():
                PREP_COUNTER.value += skipped_items  # type: ignore

    def _get_prep_data_tasks(
        self,
        prep_round: colrev.settings.PrepRound,
    ) -> list:

        prepare_data = self._load_prep_data()
        self._load_temp_prep_to_resume(prepare_data)

        self.pad = prepare_data["PAD"]
        items = prepare_data["items"]
        prep_data = []
        nr_items = len(prepare_data["items"])
        for item in items:
            prep_data.append(
                {
                    "record": colrev.record.record_prep.PrepRecord(item),
                    "nr_items": nr_items,
                    # Note : we cannot load endpoints here
                    # because pathos/multiprocessing
                    # does not support functions as parameters
                    "prep_round_package_endpoints": prep_round.prep_package_endpoints,
                    "prep_round": prep_round.name,
                }
            )
        return prep_data

    def _setup_prep_round(
        self, *, i: int, prep_round: colrev.settings.PrepRound
    ) -> None:
        """Sets up the self.prep_package_endpoints"""
        # pylint: disable=redefined-outer-name,invalid-name,global-statement
        global PREP_COUNTER
        with PREP_COUNTER.get_lock():
            PREP_COUNTER = Value("i", 0)
            PREP_COUNTER.value = 0  # type: ignore

        self.first_round = bool(i == 0)

        self.last_round = bool(
            i == len(self.review_manager.settings.prep.prep_rounds) - 1
        )

        if len(self.review_manager.settings.prep.prep_rounds) > 1:
            self.review_manager.logger.info(f"Prepare ({prep_round.name})")

        self.prep_package_endpoints: dict[str, typing.Any] = {}
        for prep_package_endpoint in prep_round.prep_package_endpoints:

            prep_class = self.package_manager.get_package_endpoint_class(
                package_type=EndpointType.prep,
                package_identifier=prep_package_endpoint["endpoint"],
            )
            self.prep_package_endpoints[prep_package_endpoint["endpoint"]] = prep_class(
                prep_operation=self, settings=prep_package_endpoint
            )

        non_available_endpoints = [
            x["endpoint"].lower()
            for x in prep_round.prep_package_endpoints
            if x["endpoint"].lower() not in self.prep_package_endpoints
        ]
        if non_available_endpoints:
            if self.review_manager.in_ci_environment():
                raise colrev_exceptions.ServiceNotAvailableException(
                    dep=f"colrev prep ({','.join(non_available_endpoints)})",
                    detailed_trace="prep not available in ci environment",
                )
            raise colrev_exceptions.ServiceNotAvailableException(
                dep="colrev prep", detailed_trace="prep not available"
            )

        for endpoint_name, endpoint in self.prep_package_endpoints.items():
            check_function = getattr(endpoint, "check_availability", None)
            if callable(check_function):
                self.review_manager.logger.debug(
                    f"Check availability of {endpoint_name}"
                )
                endpoint.check_availability(source_operation=self)  # type: ignore

    def _log_record_change_scores(
        self, *, preparation_data: list, prepared_records: list
    ) -> None:
        for previous_record_item in preparation_data:
            previous_record = previous_record_item["record"]
            prepared_record = [
                r
                for r in prepared_records
                if r[Fields.ID] == previous_record.data[Fields.ID]
            ][0]

            change = colrev.record.record_prep.PrepRecord.get_record_change_score(
                colrev.record.record_prep.PrepRecord(prepared_record),
                previous_record,
            )
            if change > 0.05:
                self.review_manager.report_logger.info(
                    f" {prepared_record[Fields.ID]} "
                    + f"Change score: {round(change, 2)}"
                )

    def _log_commit_details(self, prepared_records: list) -> None:
        nr_recs = len(
            [
                record
                for record in prepared_records
                if record[Fields.STATUS] == RecordState.md_prepared
            ]
        )

        self.review_manager.logger.info(
            "md_prepared".ljust(29)
            + f"{Colors.GREEN}{nr_recs}{Colors.END}".rjust(20, " ")
            + f" records ({nr_recs/len(prepared_records):.2%})"
        )

        nr_recs = len(
            [
                r
                for r in prepared_records
                if colrev.record.record_prep.PrepRecord(r).masterdata_is_curated()
            ]
        )

        self.review_manager.logger.info(
            "curated (✔)".ljust(29)
            + f"{Colors.GREEN}{nr_recs}{Colors.END}".rjust(20, " ")
            + f" records ({nr_recs/len(prepared_records):.2%})"
        )

        nr_recs = len(
            [
                record
                for record in prepared_records
                if record[Fields.STATUS] == RecordState.md_needs_manual_preparation
            ]
        )
        if nr_recs > 0:
            self.review_manager.logger.info(
                "md_needs_manual_preparation".ljust(29)
                + f"{Colors.ORANGE}{nr_recs}{Colors.END}".rjust(20, " ")
                + f" records ({nr_recs/len(prepared_records):.2%})"
            )

        nr_recs = len(
            [
                record
                for record in prepared_records
                if record[Fields.STATUS] == RecordState.rev_prescreen_excluded
            ]
        )
        if nr_recs > 0:
            self.review_manager.logger.info(
                "rev_prescreen_excluded".ljust(29)
                + f"{Colors.RED}{nr_recs}{Colors.END}".rjust(20, " ")
                + f" records ({nr_recs/len(prepared_records):.2%})"
            )

    def _print_startup_infos(self) -> None:
        if self.polish:
            self.review_manager.logger.info("Prep (polish mode)")
            self.review_manager.logger.info(
                "Polish mode: consider all records but prevent state transitions."
            )
        else:
            self.review_manager.logger.info("Prep")
        self.review_manager.logger.info(
            "Prep completes and corrects record metadata based on APIs and preparation rules."
        )
        self.review_manager.logger.info(
            "See https://colrev-environment.github.io/colrev/manual/metadata_retrieval/prep.html"
        )

    def _prep_packages_ram_heavy(self, prep_round: colrev.settings.PrepRound) -> bool:
        prep_pe_names = [r["endpoint"] for r in prep_round.prep_package_endpoints]
        ram_reavy = "colrev.exclude_languages" in prep_pe_names  # type: ignore
        self.review_manager.logger.info(
            "Info: The language detector requires RAM and may take longer"
        )
        return ram_reavy

    def _get_prep_pool(
        self, prep_round: colrev.settings.PrepRound
    ) -> mp.pool.ThreadPool:
        if self._prep_packages_ram_heavy(prep_round=prep_round):
            pool = Pool(mp.cpu_count() // 2)
        else:
            # Note : if we use too many CPUS,
            # a "too many open files" exception is thrown
            pool = Pool(self._cpu)
        self.review_manager.logger.info(
            "Info: ✔ = quality-assured by CoLRev community curators"
        )
        return pool

    def _create_prep_commit(
        self,
        *,
        previous_preparation_data: list,
        prepared_records: list,
    ) -> None:
        self._log_record_change_scores(
            preparation_data=previous_preparation_data,
            prepared_records=prepared_records,
        )

        self.review_manager.dataset.save_records_dict(
            {r[Fields.ID]: r for r in prepared_records}, partial=True
        )
        self._log_commit_details(prepared_records)
        self.review_manager.dataset.create_commit(
            msg="Prep: improve record metadata",
        )
        self._prep_commit_id = self.review_manager.dataset.get_repo().head.commit.hexsha
        if not self.review_manager.high_level_operation:
            print()
        self.review_manager.reset_report_logger()
        self._print_stats()

    def _post_prep(self) -> None:
        if not self.review_manager.high_level_operation:
            print()

        self.review_manager.logger.info("To validate the changes, use")
        self.review_manager.logger.info(
            f"{Colors.ORANGE}colrev validate {self._prep_commit_id}{Colors.END}"
        )
        if not self.review_manager.high_level_operation:
            print()

        self.review_manager.logger.info(
            f"{Colors.GREEN}Completed prep operation{Colors.END}"
        )
        if self.review_manager.in_ci_environment():
            print("\n\n")

    def _nothing_to_prepare_condition(self, preparation_data: list) -> bool:
        return len(preparation_data) == 0 and not self.temp_records.is_file()

    def _print_estimated_time(self, preparation_data: list) -> None:
        if self._nothing_to_prepare_condition(preparation_data):
            self.review_manager.logger.info("No records to prepare.")
            print()
            return

        self.review_manager.logger.info(f"Records to prepare: {len(preparation_data)}")
        estimated_time = len(preparation_data) * 5
        estimated_time_formatted = str(timedelta(seconds=estimated_time))
        self.review_manager.logger.info(f"Estimated time: {estimated_time_formatted}")

    @colrev.process.operation.Operation.decorate()
    def main(self, *, keep_ids: bool = False) -> None:
        """Preparation of records (main entrypoint)"""

        self._print_startup_infos()

        try:
            for i, prep_round in enumerate(
                self.review_manager.settings.prep.prep_rounds
            ):
                self._setup_prep_round(i=i, prep_round=prep_round)

                preparation_data = self._get_prep_data_tasks(prep_round)
                previous_preparation_data = deepcopy(preparation_data)

                self._print_estimated_time(preparation_data)
                if self._nothing_to_prepare_condition(preparation_data):
                    return

                if self._cpu == 1:
                    # Note: preparation_data is not turned into a list of records.
                    prepared_records = []
                    for item in preparation_data:
                        record = self.prepare(item)
                        prepared_records.append(record)
                else:
                    pool = self._get_prep_pool(prep_round)
                    prepared_records = pool.map(self.prepare, preparation_data)
                    pool.close()
                    pool.join()

                self._complete_resumed_operation(prepared_records)

                self._create_prep_commit(
                    previous_preparation_data=previous_preparation_data,
                    prepared_records=prepared_records,
                )

        except requests_ConnectionError as exc:
            if "OSError(24, 'Too many open files" in str(exc):
                raise colrev_exceptions.ServiceNotAvailableException(
                    "Too many files opened (OSError, Errno24). "
                    "To use a smaller number of parallel processes, run colrev prep --cpu 1"
                ) from exc
            raise exc

        except OSError as exc:
            if 24 == exc.errno:
                raise colrev_exceptions.ServiceNotAvailableException(
                    "Too many files opened (OSError, Errno24). "
                    "To use a smaller number of parallel processes, run colrev prep --cpu 1"
                ) from exc
            raise exc

        if not keep_ids and not self.polish:
            self.review_manager.logger.info("Set record IDs")
            self.review_manager.dataset.set_ids()
            self.review_manager.dataset.create_commit(msg="Set IDs")

        self._post_prep()



#! /usr/bin/env python
"""Checker for year-format."""
from __future__ import annotations

import re

import colrev.record.qm.quality_model
from colrev.constants import DefectCodes
from colrev.constants import Fields
from colrev.constants import FieldValues

# pylint: disable=too-few-public-methods


class YearFormatChecker:
    """The YearFormatChecker"""

    msg = DefectCodes.YEAR_FORMAT

    def __init__(
        self, quality_model: colrev.record.qm.quality_model.QualityModel
    ) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.record.Record) -> None:
        """Run the year-format checks"""

        if (
            Fields.YEAR not in record.data
            or record.ignored_defect(key=Fields.YEAR, defect=self.msg)
            or record.data[Fields.YEAR] == FieldValues.UNKNOWN
        ):
            return

        if record.masterdata_is_curated():
            return

        if not re.match(r"^\d{4}$", record.data[Fields.YEAR]):
            record.add_field_provenance_note(key=Fields.YEAR, note=self.msg)
        else:
            record.remove_field_provenance_note(key=Fields.YEAR, note=self.msg)


def register(quality_model: colrev.record.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(YearFormatChecker(quality_model))



#! /usr/bin/env python
"""Install curated CoLRev resources."""
from __future__ import annotations

from pathlib import Path

import git

import colrev.env.environment_manager
import colrev.process.operation
import colrev.record.record
from colrev.constants import Filepaths


class Resources:
    """Class for curated CoLRev resourcs (metadata repositories, annotators)"""

    # pylint: disable=too-few-public-methods

    def install_curated_resource(self, *, curated_resource: str) -> bool:
        """Install a curated resource"""

        if "http" not in curated_resource:
            curated_resource = "https://github.com/" + curated_resource
        Filepaths.CURATIONS_PATH.mkdir(exist_ok=True, parents=True)
        repo_dir = Filepaths.CURATIONS_PATH / Path(curated_resource.split("/")[-1])
        if repo_dir.is_dir():
            print(f"Repo already exists ({repo_dir})")
            return False
        print(f"Download curated resource from {curated_resource}")
        git.Repo.clone_from(curated_resource, repo_dir, depth=1)

        environment_manager = colrev.env.environment_manager.EnvironmentManager()
        if (repo_dir / Path("data/records.bib")).is_file():
            environment_manager.register_repo(repo_dir)
        elif (repo_dir / Path("README.md")).is_file():
            text = (repo_dir / Path("README.md")).read_text(encoding="utf-8")
            for line in [x for x in text.splitlines() if "colrev env --install" in x]:
                if line.endswith(curated_resource.replace("https://github.com/", "")):
                    continue
                self.install_curated_resource(
                    curated_resource=line.replace("colrev env --install ", "")
                )
        else:
            print(
                "Error: repo does not contain a "
                f"{Path('data/records.bib')} or linked repos {repo_dir}"
            )
        return True



#! /usr/bin/env python
"""Connector to doi.org (API)"""
from __future__ import annotations

import json
import re
from sqlite3 import OperationalError
from urllib.parse import unquote

import requests

import colrev.exceptions as colrev_exceptions
import colrev.package_manager.package_manager
import colrev.record.record
from colrev.constants import Fields
from colrev.constants import FieldValues
from colrev.constants import RecordState
from colrev.constants import SearchSourceHeuristicStatus
from colrev.packages.crossref.src import record_transformer

# Note: not (yet) implemented as a full search_source
# (including SearchSourceInterface, packages_endpoints.json)

# pylint: disable=too-few-public-methods


class DOIConnector:
    """Connector for the DOI.org API"""

    heuristic_status = SearchSourceHeuristicStatus.oni

    @classmethod
    def retrieve_doi_metadata(
        cls,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        record: colrev.record.record_prep.PrepRecord,
        timeout: int = 60,
    ) -> colrev.record.record.Record:
        """Retrieve the metadata from DOI.org based on a record (similarity)"""

        if Fields.DOI not in record.data:
            return record

        try:
            session = review_manager.get_cached_session()

            # for testing:
            # curl -iL -H "accept: application/vnd.citationstyles.csl+json"
            # -H "Content-Type: application/json" http://dx.doi.org/10.1111/joop.12368

            try:
                url = "http://dx.doi.org/" + record.data[Fields.DOI]
                # review_manager.logger.debug(url)
                headers = {"accept": "application/vnd.citationstyles.csl+json"}
                ret = session.request("GET", url, headers=headers, timeout=timeout)
                ret.raise_for_status()
                if ret.status_code != 200:
                    review_manager.report_logger.info(
                        f" {record.data[Fields.ID]}"
                        + "metadata for "
                        + f"doi  {record.data[Fields.DOI]} not (yet) available"
                    )
                    return record
            except OperationalError as exc:
                raise colrev_exceptions.ServiceNotAvailableException(
                    "sqlite, required for requests CachedSession "
                    "(possibly caused by concurrent operations)"
                ) from exc

            retrieved_json = json.loads(ret.text)
            language_service = colrev.env.language_service.LanguageService()
            language_service.unify_to_iso_639_3_language_codes(record=record)
            retrieved_record = record_transformer.json_to_record(item=retrieved_json)
            retrieved_record.add_provenance_all(source=url)
            record.merge(retrieved_record, default_source=url)
            record.set_masterdata_complete(
                source=url,
                masterdata_repository=review_manager.settings.is_curated_repo(),
            )
            record.set_status(RecordState.md_prepared)
            if FieldValues.RETRACTED in record.data.get("warning", ""):
                record.prescreen_exclude(reason=FieldValues.RETRACTED)
                record.remove_field(key="warning")

            if Fields.TITLE in record.data:
                record.format_if_mostly_upper(Fields.TITLE, case="sentence")

        except (requests.exceptions.RequestException,) as exc:
            print(exc)

        return record

    @classmethod
    def get_link_from_doi(
        cls,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        record: colrev.record.record.Record,
        timeout: int = 30,
    ) -> None:
        """Get the website link from DOI resolution API"""

        if Fields.DOI not in record.data:
            return

        doi_url = f"https://www.doi.org/{record.data['doi']}"

        def meta_redirect(*, content: bytes) -> str:
            content_str = content.decode("utf-8", errors="replace")

            # Ensure the content is of the expected type (contains the DOCTYPE)
            if "<!DOCTYPE HTML PUBLIC" not in content_str:
                raise TypeError

            # Regex to match the meta tag with http-equiv="REFRESH"
            meta_refresh_pattern = re.search(
                r'<meta\s+http-equiv=["\']REFRESH["\']\s+" + \
                r"content=["\']\d+;\s*url=["\']([^"\']+)["\']',
                content_str,
                re.IGNORECASE,
            )

            if meta_refresh_pattern:
                url = meta_refresh_pattern.group(1)

                # First, try to find any https URL directly in the content
                https_url_match = re.search(r'https?://[^\s"\']+', url)
                if https_url_match:
                    full_url = unquote(
                        https_url_match.group(0), encoding="utf-8", errors="replace"
                    )
                    full_url = full_url[: full_url.find("?")]
                    return str(full_url)

                # If no https URL is found, fall back to checking for a 'Redirect=' parameter
                redirect_url_match = re.search(r"Redirect=([^&]+)", url)
                if redirect_url_match:
                    # Decode and extract the full URL from the Redirect parameter
                    full_url = unquote(
                        redirect_url_match.group(1), encoding="utf-8", errors="replace"
                    )
                    full_url = full_url[: full_url.find("?")]
                    return str(full_url)

            return ""

        try:
            url = doi_url

            session = review_manager.get_cached_session()

            requests_headers = {
                "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) "
                "AppleWebKit/537.36 (KHTML, like Gecko) "
                "Chrome/39.0.2171.95 Safari/537.36"
            }
            ret = session.request(
                "GET",
                doi_url,
                headers=requests_headers,
                timeout=timeout,
            )
            if 503 == ret.status_code:
                return
            if (
                200 == ret.status_code
                and "doi.org" not in ret.url
                and "linkinghub" not in ret.url
            ):
                url = ret.url
            else:
                # follow the chain of redirects
                while meta_redirect(content=ret.content):
                    url = meta_redirect(content=ret.content)
                    ret = session.request(
                        "GET",
                        url,
                        headers=requests_headers,
                        timeout=timeout,
                    )
            record.update_field(
                key=Fields.URL,
                value=str(url.rstrip("/")),
                source=doi_url,
                keep_source_if_equal=True,
                append_edit=False,
            )
        except (requests.exceptions.RequestException, TypeError, UnicodeDecodeError):
            pass
        except OperationalError as exc:
            raise colrev_exceptions.ServiceNotAvailableException(
                "sqlite, required for requests CachedSession "
                "(possibly caused by concurrent operations)"
            ) from exc
